# -*- coding: utf-8 -*-
"""
github3.repos.repo
==================

This module contains the Repository object which is used to access the various
parts of GitHub's Repository API.

"""
from __future__ import unicode_literals

from json import dumps
from base64 import b64encode
from ..decorators import requires_auth
from ..events import Event
from ..git import Blob, Commit, Reference, Tag, Tree
from ..issues import issue_params, Issue
from ..issues.event import IssueEvent
from ..issues.label import Label
from ..issues.milestone import Milestone
from ..models import GitHubCore
from ..notifications import Subscription, Thread
from ..pulls import PullRequest
from .branch import Branch
from .comment import RepoComment
from .commit import RepoCommit
from .comparison import Comparison
from .contents import Contents, validate_commmitter
from .deployment import Deployment
from .hook import Hook
from .issue_import import ImportedIssue
from ..licenses import License
from .pages import PagesBuild, PagesInfo
from .status import Status
from .stats import ContributorStats
from .release import Release, Asset
from .tag import RepoTag
from ..users import User, Key
from ..utils import stream_response_to_file, timestamp_parameter
from uritemplate import URITemplate


class Repository(GitHubCore):

    """The :class:`Repository <Repository>` object.

    It represents how GitHub sends information about repositories.

    Two repository instances can be checked like so::

        r1 == r2
        r1 != r2

    And is equivalent to::

        r1.id == r2.id
        r1.id != r2.id

    See also: http://developer.github.com/v3/repos/

    """

    STAR_HEADERS = {
        'Accept': 'application/vnd.github.v3.star+json'
    }

    def _update_attributes(self, repo):
        #: URL used to clone via HTTPS.
        self.clone_url = repo.get('clone_url', '')
        #: ``datetime`` object representing when the Repository was created.
        self.created_at = self._strptime(repo.get('created_at'))
        #: Description of the repository.
        self.description = repo.get('description', '')

        #: The number of forks of this repository.
        self.forks_count = repo.get('forks_count')
        #: The number of forks of this repository. For backward compatibility
        self.fork_count = self.forks_count

        #: Is this repository a fork?
        self.fork = repo.get('fork')

        #: Full name as login/name
        self.full_name = repo.get('full_name', '')

        # Clone url using git, e.g. git://github.com/sigmavirus24/github3.py
        #: Plain git url for an anonymous clone.
        self.git_url = repo.get('git_url', '')
        #: Whether or not this repository has downloads enabled
        self.has_downloads = repo.get('has_downloads')
        #: Whether or not this repository has an issue tracker
        self.has_issues = repo.get('has_issues')
        #: Whether or not this repository has the wiki enabled
        self.has_wiki = repo.get('has_wiki')

        # e.g. https://sigmavirus24.github.com/github3.py
        #: URL of the home page for the project.
        self.homepage = repo.get('homepage', '')

        #: URL of the pure diff of the pull request
        self.diff_url = repo.get('diff_url', '')

        #: URL of the pure patch of the pull request
        self.patch_url = repo.get('patch_url', '')

        #: API URL of the issue representation of this Pull Request
        self.issue_url = repo.get('issue_url', '')

        # e.g. https://github.com/sigmavirus24/github3.py
        #: URL of the project at GitHub.
        self.html_url = repo.get('html_url', '')
        #: Unique id of the repository.
        self.id = repo.get('id', 0)
        #: Language property.
        self.language = repo.get('language', '')
        #: Mirror property.
        self.mirror_url = repo.get('mirror_url', '')

        # Repository name, e.g. github3.py
        #: Name of the repository.
        self.name = repo.get('name', '')

        #: Number of open issues on the repository. DEPRECATED
        self.open_issues = repo.get('open_issues', 0)

        #: Number of open issues on the repository
        self.open_issues_count = repo.get('open_issues_count')

        # Repository owner's name
        #: :class:`User <github3.users.User>` object representing the
        #: repository owner.
        self.owner = User(repo.get('owner', {}), self)

        #: Is this repository private?
        self.private = repo.get('private')

        #: Permissions for this repository
        self.permissions = repo.get('permissions')

        #: ``datetime`` object representing the last time commits were pushed
        #: to the repository.
        self.pushed_at = self._strptime(repo.get('pushed_at'))
        #: Size of the repository.
        self.size = repo.get('size', 0)

        # The number of stargazers
        #: Number of users who starred the repository
        self.stargazers_count = repo.get('stargazers_count', 0)

        #: ``datetime`` object representing when the repository was starred
        self.starred_at = self._strptime(repo.get('starred_at'))

        # SSH url e.g. git@github.com/sigmavirus24/github3.py
        #: URL to clone the repository via SSH.
        self.ssh_url = repo.get('ssh_url', '')
        #: If it exists, url to clone the repository via SVN.
        self.svn_url = repo.get('svn_url', '')
        #: ``datetime`` object representing the last time the repository was
        #: updated.
        self.updated_at = self._strptime(repo.get('updated_at'))
        self._api = repo.get('url', '')

        # The number of watchers
        #: Number of users watching the repository.
        self.watchers = repo.get('watchers', 0)

        #: Parent of this fork, if it exists :class:`Repository`
        self.source = repo.get('source')
        if self.source:
            self.source = Repository(self.source, self)

        #: Parent of this fork, if it exists :class:`Repository`
        self.parent = repo.get('parent')
        if self.parent:
            self.parent = Repository(self.parent, self)

        #: default branch for the repository
        self.default_branch = repo.get('default_branch', '')

        #: master (default) branch for the repository
        self.master_branch = repo.get('master_branch', '')

        #: Teams url (not a template)
        self.teams_url = repo.get('teams_url', '')

        #: Hooks url (not a template)
        self.hooks_url = repo.get('hooks_url', '')

        #: Events url (not a template)
        self.events_url = repo.get('events_url', '')

        #: Tags url (not a template)
        self.tags_url = repo.get('tags_url', '')

        #: Languages url (not a template)
        self.languages_url = repo.get('languages_url', '')

        #: Stargazers url (not a template)
        self.stargazers_url = repo.get('stargazers_url', '')

        #: Contributors url (not a template)
        self.contributors_url = repo.get('contributors_url', '')

        #: Subscribers url (not a template)
        self.subscribers_url = repo.get('subscribers_url', '')

        #: Subscription url (not a template)
        self.subscription_url = repo.get('subscription_url', '')

        #: Merges url (not a template)
        self.merges_url = repo.get('merges_url', '')

        #: Downloads url (not a template)
        self.download_url = repo.get('downloads_url', '')

        # Template URLS
        ie_url_t = repo.get('issue_events_url')
        #: Issue events URL Template. Expand with ``number``
        self.issue_events_urlt = URITemplate(ie_url_t) if ie_url_t else None

        assignees = repo.get('assignees_url')
        #: Assignees URL Template. Expand with ``user``
        self.assignees_urlt = URITemplate(assignees) if assignees else None

        branches = repo.get('branches_url')
        #: Branches URL Template. Expand with ``branch``
        self.branches_urlt = URITemplate(branches) if branches else None

        blobs = repo.get('blobs_url')
        #: Blobs URL Template. Expand with ``sha``
        self.blobs_urlt = URITemplate(blobs) if blobs else None

        git_tags = repo.get('git_tags_url')
        #: Git tags URL Template. Expand with ``sha``
        self.git_tags_urlt = URITemplate(git_tags) if git_tags else None

        git_refs = repo.get('git_refs_url')
        #: Git refs URL Template. Expand with ``sha``
        self.git_refs_urlt = URITemplate(git_refs) if git_refs else None

        trees = repo.get('trees_url')
        #: Trres URL Template. Expand with ``sha``
        self.trees_urlt = URITemplate(trees) if trees else None

        statuses = repo.get('statuses_url')
        #: Statuses URL Template. Expand with ``sha``
        self.statuses_urlt = URITemplate(statuses) if statuses else None

        commits = repo.get('commits_url')
        #: Commits URL Template. Expand with ``sha``
        self.commits_urlt = URITemplate(commits) if commits else None

        commits = repo.get('git_commits_url')
        #: Git commits URL Template. Expand with ``sha``
        self.git_commits_urlt = URITemplate(commits) if commits else None

        comments = repo.get('comments_url')
        #: Comments URL Template. Expand with ``number``
        self.comments_urlt = URITemplate(comments) if comments else None

        comments = repo.get('review_comments_url')
        #: Pull Request Review Comments URL
        self.review_comments_url = URITemplate(comments) if comments else None

        comments = repo.get('review_comment_url')
        #: Pull Request Review Comments URL Template. Expand with ``number``
        self.review_comment_urlt = URITemplate(comments) if comments else None

        comments = repo.get('issue_comment_url')
        #: Issue comment URL Template. Expand with ``number``
        self.issue_comment_urlt = URITemplate(comments) if comments else None

        contents = repo.get('contents_url')
        #: Contents URL Template. Expand with ``path``
        self.contents_urlt = URITemplate(contents) if contents else None

        compare = repo.get('compare_url')
        #: Comparison URL Template. Expand with ``base`` and ``head``
        self.compare_urlt = URITemplate(compare) if compare else None

        archive = repo.get('archive_url')
        #: Archive URL Template. Expand with ``archive_format`` and ``ref``
        self.archive_urlt = URITemplate(archive) if archive else None

        issues = repo.get('issues_url')
        #: Issues URL Template. Expand with ``number``
        self.issues_urlt = URITemplate(issues) if issues else None

        pulls = repo.get('pulls_url')
        #: Pull Requests URL Template. Expand with ``number``
        self.pulls_urlt = URITemplate(pulls) if issues else None

        miles = repo.get('milestones_url')
        #: Milestones URL Template. Expand with ``number``
        self.milestones_urlt = URITemplate(miles) if miles else None

        notif = repo.get('notifications_url')
        #: Notifications URL Template. Expand with ``since``, ``all``,
        #: ``participating``
        self.notifications_urlt = URITemplate(notif) if notif else None

        labels = repo.get('labels_url')
        #: Labels URL Template. Expand with ``name``
        self.labels_urlt = URITemplate(labels) if labels else None

    def _repr(self):
        return '<Repository [{0}]>'.format(self)

    def __str__(self):
        return self.full_name

    def _create_pull(self, data):
        self._remove_none(data)
        json = None
        if data:
            url = self._build_url('pulls', base_url=self._api)
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(PullRequest, json)

    @requires_auth
    def add_collaborator(self, username):
        """Add ``username`` as a collaborator to a repository.

        :param username: (required), username of the user
        :type username: str or :class:`User <github3.users.User>`
        :returns: bool -- True if successful, False otherwise
        """
        if not username:
            return False
        url = self._build_url('collaborators', str(username),
                              base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    def archive(self, format, path='', ref='master'):
        """Get the tarball or zipball archive for this repo at ref.

        See: http://developer.github.com/v3/repos/contents/#get-archive-link

        :param str format: (required), accepted values: ('tarball',
            'zipball')
        :param path: (optional), path where the file should be saved
            to, default is the filename provided in the headers and will be
            written in the current directory.
            it can take a file-like object as well
        :type path: str, file
        :param str ref: (optional)
        :returns: bool -- True if successful, False otherwise

        """
        resp = None
        if format in ('tarball', 'zipball'):
            url = self._build_url(format, ref, base_url=self._api)
            resp = self._get(url, allow_redirects=True, stream=True)

        if resp and self._boolean(resp, 200, 404):
            stream_response_to_file(resp, path)
            return True
        return False

    def asset(self, id):
        """Return a single asset.

        :param int id: (required), id of the asset
        :returns: :class:`Asset <github3.repos.release.Asset>`
        """
        data = None
        if int(id) > 0:
            url = self._build_url('releases', 'assets', str(id),
                                  base_url=self._api)
            data = self._json(self._get(url, headers=Release.CUSTOM_HEADERS),
                              200)
        return self._instance_or_null(Asset, data)

    def assignees(self, number=-1, etag=None):
        r"""Iterate over all assignees to which an issue may be assigned.

        :param int number: (optional), number of assignees to return. Default:
            -1 returns all available assignees
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <github3.users.User>`\ s
        """
        url = self._build_url('assignees', base_url=self._api)
        return self._iter(int(number), url, User, etag=etag)

    def blob(self, sha):
        """Get the blob indicated by ``sha``.

        :param str sha: (required), sha of the blob
        :returns: :class:`Blob <github3.git.Blob>` if successful, otherwise
            None
        """
        url = self._build_url('git', 'blobs', sha, base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(Blob, json)

    def branch(self, name):
        """Get the branch ``name`` of this repository.

        :param str name: (required), branch name
        :type name: str
        :returns: :class:`Branch <github3.repos.branch.Branch>`
        """
        json = None
        if name:
            url = self._build_url('branches', name, base_url=self._api)
            json = self._json(self._get(url, headers=Branch.PREVIEW_HEADERS),
                              200)
        return self._instance_or_null(Branch, json)

    def branches(self, number=-1, protected=False, etag=None):
        r"""Iterate over the branches in this repository.

        :param int number: (optional), number of branches to return. Default:
            -1 returns all branches
        :param bool protected: (optional), True lists only protected branches.
            Default: False
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`Branch <github3.repos.branch.Branch>`\ es
        """
        url = self._build_url('branches', base_url=self._api)
        params = {'protected': '1'} if protected else None
        return self._iter(int(number), url, Branch, params, etag=etag,
                          headers=Branch.PREVIEW_HEADERS)

    def code_frequency(self, number=-1, etag=None):
        """Iterate over the code frequency per week.

        Returns a weekly aggregate of the number of additions and deletions
        pushed to this repository.

        :param int number: (optional), number of weeks to return. Default: -1
            returns all weeks
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of lists ``[seconds_from_epoch, additions,
            deletions]``

        .. note:: All statistics methods may return a 202. On those occasions,
                  you will not receive any objects. You should store your
                  iterator and check the new ``last_status`` attribute. If it
                  is a 202 you should wait before re-requesting.

        .. versionadded:: 0.7

        """
        url = self._build_url('stats', 'code_frequency', base_url=self._api)
        return self._iter(int(number), url, list, etag=etag)

    def collaborators(self, number=-1, etag=None):
        r"""Iterate over the collaborators of this repository.

        :param int number: (optional), number of collaborators to return.
            Default: -1 returns all comments
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <github3.users.User>`\ s
        """
        url = self._build_url('collaborators', base_url=self._api)
        return self._iter(int(number), url, User, etag=etag)

    def comments(self, number=-1, etag=None):
        r"""Iterate over comments on all commits in the repository.

        :param int number: (optional), number of comments to return. Default:
            -1 returns all comments
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`RepoComment <github3.repos.comment.RepoComment>`\ s
        """
        url = self._build_url('comments', base_url=self._api)
        return self._iter(int(number), url, RepoComment, etag=etag)

    def commit(self, sha):
        """Get a single (repo) commit.

        See :func:`git_commit` for the Git Data Commit.

        :param str sha: (required), sha of the commit
        :returns: :class:`RepoCommit <github3.repos.commit.RepoCommit>` if
            successful, otherwise None
        """
        url = self._build_url('commits', sha, base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(RepoCommit, json)

    def commit_activity(self, number=-1, etag=None):
        """Iterate over last year of commit activity by week.

        See: http://developer.github.com/v3/repos/statistics/

        .. note:: All statistics methods may return a 202. On those occasions,
                  you will not receive any objects. You should store your
                  iterator and check the new ``last_status`` attribute. If it
                  is a 202 you should wait before re-requesting.

        .. versionadded:: 0.7

        :param int number: (optional), number of weeks to return. Default -1
            will return all of the weeks.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of dictionaries
        """
        url = self._build_url('stats', 'commit_activity', base_url=self._api)
        return self._iter(int(number), url, dict, etag=etag)

    def commit_comment(self, comment_id):
        """Get a single commit comment.

        :param int comment_id: (required), id of the comment used by GitHub
        :returns: :class:`RepoComment <github3.repos.comment.RepoComment>` if
            successful, otherwise None
        """
        url = self._build_url('comments', str(comment_id), base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(RepoComment, json)

    def commits(self, sha=None, path=None, author=None, number=-1, etag=None,
                since=None, until=None):
        r"""Iterate over commits in this repository.

        :param str sha: (optional), sha or branch to start listing commits
            from
        :param str path: (optional), commits containing this path will be
            listed
        :param str author: (optional), GitHub login, real name, or email to
            filter commits by (using commit author)
        :param int number: (optional), number of comments to return. Default:
            -1 returns all comments
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :param since: (optional), Only commits after this date will
            be returned. This can be a ``datetime`` or an ``ISO8601`` formatted
            date string.
        :type since: datetime or string
        :param until: (optional), Only commits before this date will
            be returned. This can be a ``datetime`` or an ``ISO8601`` formatted
            date string.
        :type until: datetime or string

        :returns: generator of
            :class:`RepoCommit <github3.repos.commit.RepoCommit>`\ s
        """
        params = {'sha': sha, 'path': path, 'author': author,
                  'since': timestamp_parameter(since),
                  'until': timestamp_parameter(until)}

        self._remove_none(params)
        url = self._build_url('commits', base_url=self._api)
        return self._iter(int(number), url, RepoCommit, params, etag)

    def compare_commits(self, base, head):
        """Compare two commits.

        :param str base: (required), base for the comparison
        :param str head: (required), compare this against base
        :returns: :class:`Comparison <github3.repos.comparison.Comparison>` if
            successful, else None
        """
        url = self._build_url('compare', base + '...' + head,
                              base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(Comparison, json)

    def contributor_statistics(self, number=-1, etag=None):
        """Iterate over the contributors list.

        See also: http://developer.github.com/v3/repos/statistics/

        .. note:: All statistics methods may return a 202. On those occasions,
                  you will not receive any objects. You should store your
                  iterator and check the new ``last_status`` attribute. If it
                  is a 202 you should wait before re-requesting.

        .. versionadded:: 0.7

        :param int number: (optional), number of weeks to return. Default -1
            will return all of the weeks.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`ContributorStats <github3.repos.stats.ContributorStats>`
        """
        url = self._build_url('stats', 'contributors', base_url=self._api)
        return self._iter(int(number), url, ContributorStats, etag=etag)

    def contributors(self, anon=False, number=-1, etag=None):
        r"""Iterate over the contributors to this repository.

        :param bool anon: (optional), True lists anonymous contributors as
            well
        :param int number: (optional), number of contributors to return.
            Default: -1 returns all contributors
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <github3.users.User>`\ s
        """
        url = self._build_url('contributors', base_url=self._api)
        params = {}
        if anon:
            params = {'anon': 'true'}
        return self._iter(int(number), url, User, params, etag)

    @requires_auth
    def create_blob(self, content, encoding):
        """Create a blob with ``content``.

        :param str content: (required), content of the blob
        :param str encoding: (required), ('base64', 'utf-8')
        :returns: string of the SHA returned
        """
        sha = ''
        if encoding in ('base64', 'utf-8'):
            url = self._build_url('git', 'blobs', base_url=self._api)
            data = {'content': content, 'encoding': encoding}
            json = self._json(self._post(url, data=data), 201)
            if json:
                sha = json.get('sha')
        return sha

    @requires_auth
    def create_comment(self, body, sha, path=None, position=None, line=1):
        """Create a comment on a commit.

        :param str body: (required), body of the message
        :param str sha: (required), commit id
        :param str path: (optional), relative path of the file to comment
            on
        :param str position: (optional), line index in the diff to comment on
        :param int line: (optional), line number of the file to comment on,
            default: 1
        :returns: :class:`RepoComment <github3.repos.comment.RepoComment>` if
            successful, otherwise None

        """
        json = None
        if body and sha and (line and int(line) > 0):
            data = {'body': body, 'line': line, 'path': path,
                    'position': position}
            self._remove_none(data)
            url = self._build_url('commits', sha, 'comments',
                                  base_url=self._api)
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(RepoComment, json)

    @requires_auth
    def create_commit(self, message, tree, parents, author=None,
                      committer=None):
        """Create a commit on this repository.

        :param str message: (required), commit message
        :param str tree: (required), SHA of the tree object this
            commit points to
        :param list parents: (required), SHAs of the commits that were parents
            of this commit. If empty, the commit will be written as the root
            commit.  Even if there is only one parent, this should be an
            array.
        :param dict author: (optional), if omitted, GitHub will
            use the authenticated user's credentials and the current
            time. Format: {'name': 'Committer Name', 'email':
            'name@example.com', 'date': 'YYYY-MM-DDTHH:MM:SS+HH:00'}
        :param dict committer: (optional), if ommitted, GitHub will use the
            author parameters. Should be the same format as the author
            parameter.
        :returns: :class:`Commit <github3.git.Commit>` if successful, else
            None
        """
        json = None
        if message and tree and isinstance(parents, list):
            url = self._build_url('git', 'commits', base_url=self._api)
            data = {'message': message, 'tree': tree, 'parents': parents,
                    'author': author, 'committer': committer}
            self._remove_none(data)
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(Commit, json)

    @requires_auth
    def create_deployment(self, ref, force=False, payload='',
                          auto_merge=False, description='', environment=None):
        """Create a deployment.

        :param str ref: (required), The ref to deploy. This can be a branch,
            tag, or sha.
        :param bool force: Optional parameter to bypass any ahead/behind
            checks or commit status checks. Default: False
        :param str payload: Optional JSON payload with extra information about
            the deployment. Default: ""
        :param bool auto_merge: Optional parameter to merge the default branch
            into the requested deployment branch if necessary. Default: False
        :param str description: Optional short description. Default: ""
        :param str environment: Optional name for the target deployment
            environment (e.g., production, staging, qa). Default: "production"
        :returns: :class:`Deployment <github3.repos.deployment.Deployment>`
        """
        json = None
        if ref:
            url = self._build_url('deployments', base_url=self._api)
            data = {'ref': ref, 'force': force, 'payload': payload,
                    'auto_merge': auto_merge, 'description': description,
                    'environment': environment}
            self._remove_none(data)
            json = self._json(self._post(url, data=data),
                              201)
        return self._instance_or_null(Deployment, json)

    @requires_auth
    def create_file(self, path, message, content, branch=None,
                    committer=None, author=None):
        """Create a file in this repository.

        See also: http://developer.github.com/v3/repos/contents/#create-a-file

        :param str path: (required), path of the file in the repository
        :param str message: (required), commit message
        :param bytes content: (required), the actual data in the file
        :param str branch: (optional), branch to create the commit on.
            Defaults to the default branch of the repository
        :param dict committer: (optional), if no information is given the
            authenticated user's information will be used. You must specify
            both a name and email.
        :param dict author: (optional), if omitted this will be filled in with
            committer information. If passed, you must specify both a name and
            email.
        :returns: {
            'content': :class:`Contents <github3.repos.contents.Contents>`:,
            'commit': :class:`Commit <github3.git.Commit>`}

        """
        if content and not isinstance(content, bytes):
            raise ValueError(  # (No coverage)
                'content must be a bytes object')  # (No coverage)

        json = None
        if path and message and content:
            url = self._build_url('contents', path, base_url=self._api)
            content = b64encode(content).decode('utf-8')
            data = {'message': message, 'content': content, 'branch': branch,
                    'committer': validate_commmitter(committer),
                    'author': validate_commmitter(author)}
            self._remove_none(data)
            json = self._json(self._put(url, data=dumps(data)), 201)
            if json and 'content' in json and 'commit' in json:
                json['content'] = Contents(json['content'], self)
                json['commit'] = Commit(json['commit'], self)
        return json

    @requires_auth
    def create_fork(self, organization=None):
        """Create a fork of this repository.

        :param str organization: (required), login for organization to create
            the fork under
        :returns: :class:`Repository <Repository>` if successful, else None
        """
        url = self._build_url('forks', base_url=self._api)
        if organization:
            resp = self._post(url, data={'organization': organization})
        else:
            resp = self._post(url)

        json = self._json(resp, 202)
        return self._instance_or_null(Repository, json)

    @requires_auth
    def create_hook(self, name, config, events=['push'], active=True):
        """Create a hook on this repository.

        :param str name: (required), name of the hook
        :param dict config: (required), key-value pairs which act as settings
            for this hook
        :param list events: (optional), events the hook is triggered for
        :param bool active: (optional), whether the hook is actually
            triggered
        :returns: :class:`Hook <github3.repos.hook.Hook>` if successful,
            otherwise None
        """
        json = None
        if name and config and isinstance(config, dict):
            url = self._build_url('hooks', base_url=self._api)
            data = {'name': name, 'config': config, 'events': events,
                    'active': active}
            json = self._json(self._post(url, data=data), 201)
        return Hook(json, self) if json else None

    @requires_auth
    def create_issue(self,
                     title,
                     body=None,
                     assignee=None,
                     milestone=None,
                     labels=None):
        """Create an issue on this repository.

        :param str title: (required), title of the issue
        :param str body: (optional), body of the issue
        :param str assignee: (optional), login of the user to assign the
            issue to
        :param int milestone: (optional), id number of the milestone to
            attribute this issue to (e.g. ``m`` is a :class:`Milestone
            <github3.issues.milestone.Milestone>` object, ``m.number`` is
            what you pass here.)
        :param labels: (optional), labels to apply to this
            issue
        :type labels: list of strings
        :returns: :class:`Issue <github3.issues.issue.Issue>` if successful,
            otherwise None
        """
        issue = {'title': title, 'body': body, 'assignee': assignee,
                 'milestone': milestone, 'labels': labels}
        self._remove_none(issue)
        json = None

        if issue:
            url = self._build_url('issues', base_url=self._api)
            json = self._json(self._post(url, data=issue), 201)

        return self._instance_or_null(Issue, json)

    @requires_auth
    def create_key(self, title, key, read_only=False):
        """Create a deploy key.

        :param str title: (required), title of key
        :param str key: (required), key text
        :param bool read_only: (optional), restrict key access to read-only,
            default is False
        :returns: :class:`Key <github3.users.Key>` if successful, else None
        """
        json = None
        if title and key:
            data = {'title': title, 'key': key, 'read_only': read_only}
            url = self._build_url('keys', base_url=self._api)
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(Key, json)

    @requires_auth
    def create_label(self, name, color):
        """Create a label for this repository.

        :param str name: (required), name to give to the label
        :param str color: (required), value of the color to assign to the
            label, e.g., '#fafafa' or 'fafafa' (the latter is what is sent)
        :returns: :class:`Label <github3.issues.label.Label>` if successful,
            else None
        """
        json = None
        if name and color:
            data = {'name': name, 'color': color.strip('#')}
            url = self._build_url('labels', base_url=self._api)
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(Label, json)

    @requires_auth
    def create_milestone(self, title, state=None, description=None,
                         due_on=None):
        """Create a milestone for this repository.

        :param str title: (required), title of the milestone
        :param str state: (optional), state of the milestone, accepted
            values: ('open', 'closed'), default: 'open'
        :param str description: (optional), description of the milestone
        :param str due_on: (optional), ISO 8601 formatted due date
        :returns: :class:`Milestone <github3.issues.milestone.Milestone>` if
            successful, otherwise None
        """
        url = self._build_url('milestones', base_url=self._api)
        if state not in ('open', 'closed'):
            state = None
        data = {'title': title, 'state': state,
                'description': description, 'due_on': due_on}
        self._remove_none(data)
        json = None
        if data:
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(Milestone, json)

    @requires_auth
    def create_pull(self, title, base, head, body=None):
        """Create a pull request of ``head`` onto ``base`` branch in this repo.

        :param str title: (required)
        :param str base: (required), e.g., 'master'
        :param str head: (required), e.g., 'username:branch'
        :param str body: (optional), markdown formatted description
        :returns: :class:`PullRequest <github3.pulls.PullRequest>` if
            successful, else None
        """
        data = {'title': title, 'body': body, 'base': base,
                'head': head}
        return self._create_pull(data)

    @requires_auth
    def create_pull_from_issue(self, issue, base, head):
        """Create a pull request from issue #``issue``.

        :param int issue: (required), issue number
        :param str base: (required), e.g., 'master'
        :param str head: (required), e.g., 'username:branch'
        :returns: :class:`PullRequest <github3.pulls.PullRequest>` if
            successful, else None
        """
        if int(issue) > 0:
            data = {'issue': issue, 'base': base, 'head': head}
            return self._create_pull(data)
        return None

    @requires_auth
    def create_ref(self, ref, sha):
        """Create a reference in this repository.

        :param str ref: (required), fully qualified name of the reference,
            e.g. ``refs/heads/master``. If it doesn't start with ``refs`` and
            contain at least two slashes, GitHub's API will reject it.
        :param str sha: (required), SHA1 value to set the reference to
        :returns: :class:`Reference <github3.git.Reference>` if successful
            else None
        """
        json = None
        if ref and ref.startswith('refs') and ref.count('/') >= 2 and sha:
            data = {'ref': ref, 'sha': sha}
            url = self._build_url('git', 'refs', base_url=self._api)
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(Reference, json)

    @requires_auth
    def create_release(self, tag_name, target_commitish=None, name=None,
                       body=None, draft=False, prerelease=False):
        """Create a release for this repository.

        :param str tag_name: (required), name to give to the tag
        :param str target_commitish: (optional), vague concept of a target,
            either a SHA or a branch name.
        :param str name: (optional), name of the release
        :param str body: (optional), description of the release
        :param bool draft: (optional), whether this release is a draft or not
        :param bool prerelease: (optional), whether this is a prerelease or
            not
        :returns: :class:`Release <github3.repos.release.Release>`
        """
        data = {'tag_name': str(tag_name),
                'target_commitish': target_commitish,
                'name': name,
                'body': body,
                'draft': draft,
                'prerelease': prerelease
                }
        self._remove_none(data)

        url = self._build_url('releases', base_url=self._api)
        json = self._json(self._post(
            url, data=data, headers=Release.CUSTOM_HEADERS
            ), 201)
        return self._instance_or_null(Release, json)

    @requires_auth
    def create_status(self, sha, state, target_url=None, description=None,
                      context='default'):
        """Create a status object on a commit.

        :param str sha: (required), SHA of the commit to create the status on
        :param str state: (required), state of the test; only the following
            are accepted: 'pending', 'success', 'error', 'failure'
        :param str target_url: (optional), URL to associate with this status.
        :param str description: (optional), short description of the status
        :param str context: (optional), A string label to differentiate this
            status from the status of other systems
        :returns: the status created if successful
        :rtype: :class:`~github3.repos.status.Status`
        """
        json = None
        if sha and state:
            data = {'state': state, 'target_url': target_url,
                    'description': description, 'context': context}
            url = self._build_url('statuses', sha, base_url=self._api)
            self._remove_none(data)
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(Status, json)

    @requires_auth
    def create_tag(self, tag, message, sha, obj_type, tagger,
                   lightweight=False):
        """Create a tag in this repository.

        By default, this method creates an annotated tag. If you wish to
        create a lightweight tag instead, pass ``lightweight=True``.

        If you are creating an annotated tag, this method makes **2 calls** to
        the API:

        1. Creates the tag object
        2. Creates the reference for the tag

        This behaviour is required by the GitHub API.

        :param str tag: (required), name of the tag
        :param str message: (required), tag message
        :param str sha: (required), SHA of the git object this is tagging
        :param str obj_type: (required), type of object being tagged, e.g.,
            'commit', 'tree', 'blob'
        :param dict tagger: (required), containing the name, email of the
            tagger and the date it was tagged
        :param bool lightweight: (optional), if False, create an annotated
            tag, otherwise create a lightweight tag (a Reference).
        :returns: If lightweight == False: :class:`Tag <github3.git.Tag>` if
            successful, else None. If lightweight == True: :class:`Reference
            <github3.git.Reference>`
        """
        if lightweight and tag and sha:
            return self.create_ref('refs/tags/' + tag, sha)

        json = None
        if tag and message and sha and obj_type and len(tagger) == 3:
            data = {'tag': tag, 'message': message, 'object': sha,
                    'type': obj_type, 'tagger': tagger}
            url = self._build_url('git', 'tags', base_url=self._api)
            json = self._json(self._post(url, data=data), 201)
            if json:
                self.create_ref('refs/tags/' + tag, json.get('sha'))
        return self._instance_or_null(Tag, json)

    @requires_auth
    def create_tree(self, tree, base_tree=None):
        """Create a tree on this repository.

        :param list tree: (required), specifies the tree structure.
            Format: [{'path': 'path/file', 'mode':
            'filemode', 'type': 'blob or tree', 'sha': '44bfc6d...'}]
        :param str base_tree: (optional), SHA1 of the tree you want
            to update with new data
        :returns: :class:`Tree <github3.git.Tree>` if successful, else None
        """
        json = None
        if tree and isinstance(tree, list):
            data = {'tree': tree}
            if base_tree:
                data['base_tree'] = base_tree
            url = self._build_url('git', 'trees', base_url=self._api)
            json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(Tree, json)

    @requires_auth
    def delete(self):
        """Delete this repository.

        :returns: bool -- True if successful, False otherwise
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def delete_key(self, key_id):
        """Delete the key with the specified id from your deploy keys list.

        :returns: bool -- True if successful, False otherwise
        """
        if int(key_id) <= 0:
            return False
        url = self._build_url('keys', str(key_id), base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    @requires_auth
    def delete_subscription(self):
        """Delete the user's subscription to this repository.

        :returns: bool
        """
        url = self._build_url('subscription', base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    def deployment(self, id):
        """Retrieve the deployment identified by ``id``.

        :param int id: (required), id for deployments.
        :returns: :class:`~github3.repos.deployment.Deployment`
        """
        json = None
        if int(id) > 0:
            url = self._build_url('deployments', str(id), base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Deployment, json)

    def deployments(self, number=-1, etag=None):
        r"""Iterate over deployments for this repository.

        :param int number: (optional), number of deployments to return.
            Default: -1, returns all available deployments
        :param str etag: (optional), ETag from a previous request for all
            deployments
        :returns: generator of
            :class:`Deployment <github3.repos.deployment.Deployment>`\ s
        """
        url = self._build_url('deployments', base_url=self._api)
        i = self._iter(int(number), url, Deployment, etag=etag)
        return i

    def directory_contents(self, directory_path, ref=None, return_as=list):
        """Get the contents of each file in ``directory_path``.

        If the path provided is actually a directory, you will receive a
        list back of the form::

            [('filename.md', Contents(...)),
             ('github.py', Contents(...)),
             # ...
             ('fiz.py', Contents(...))]

        You can either then transform it into a dictionary::

            contents = dict(repo.directory_contents('path/to/dir/'))

        Or you can use the ``return_as`` parameter to have it return a
        dictionary for you::

            contents = repo.directory_contents('path/to/dir/', return_as=dict)

        :param str path: (required), path to file, e.g.
            github3/repos/repo.py
        :param str ref: (optional), the string name of a commit/branch/tag.
            Default: master
        :param return_as: (optional), how to return the directory's contents.
            Default: :class:`list`
        :returns: list of tuples of the filename and the Contents returned
        :rtype: list((str, :class:`~github3.repos.contents.Contents`))
        """
        url = self._build_url('contents', directory_path, base_url=self._api)
        json = self._json(self._get(url, params={'ref': ref}), 200) or []
        return return_as((j.get('name'), Contents(j, self)) for j in json)

    @requires_auth
    def edit(self, name, description=None, homepage=None, private=None,
             has_issues=None, has_wiki=None, has_downloads=None,
             default_branch=None):
        """Edit this repository.

        :param str name: (required), name of the repository
        :param str description: (optional), If not ``None``, change the
            description for this repository. API default: ``None`` - leave
            value unchanged.
        :param str homepage: (optional), If not ``None``, change the homepage
            for this repository. API default: ``None`` - leave value unchanged.
        :param bool private: (optional), If ``True``, make the repository
            private. If ``False``, make the repository public. API default:
            ``None`` - leave value unchanged.
        :param bool has_issues: (optional), If ``True``, enable issues for
            this repository. If ``False``, disable issues for this repository.
            API default: ``None`` - leave value unchanged.
        :param bool has_wiki: (optional), If ``True``, enable the wiki for
            this repository. If ``False``, disable the wiki for this
            repository. API default: ``None`` - leave value unchanged.
        :param bool has_downloads: (optional), If ``True``, enable downloads
            for this repository. If ``False``, disable downloads for this
            repository. API default: ``None`` - leave value unchanged.
        :param str default_branch: (optional), If not ``None``, change the
            default branch for this repository. API default: ``None`` - leave
            value unchanged.
        :returns: bool -- True if successful, False otherwise
        """
        edit = {'name': name, 'description': description, 'homepage': homepage,
                'private': private, 'has_issues': has_issues,
                'has_wiki': has_wiki, 'has_downloads': has_downloads,
                'default_branch': default_branch}
        self._remove_none(edit)
        json = None
        if edit:
            json = self._json(self._patch(self._api, data=dumps(edit)), 200)
            self._update_attributes(json)
            return True
        return False

    def events(self, number=-1, etag=None):
        r"""Iterate over events on this repository.

        :param int number: (optional), number of events to return. Default: -1
            returns all available events
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Event <github3.events.Event>`\ s
        """
        url = self._build_url('events', base_url=self._api)
        return self._iter(int(number), url, Event, etag=etag)

    def file_contents(self, path, ref=None):
        """Get the contents of the file pointed to by ``path``.

        :param str path: (required), path to file, e.g.
            github3/repos/repo.py
        :param str ref: (optional), the string name of a commit/branch/tag.
            Default: master
        :returns: the contents of the file requested
        :rtype: :class:`~github3.repos.contents.Contents`
        """
        url = self._build_url('contents', path, base_url=self._api)
        json = self._json(self._get(url, params={'ref': ref}), 200)
        return self._instance_or_null(Contents, json)

    def forks(self, sort='', number=-1, etag=None):
        """Iterate over forks of this repository.

        :param str sort: (optional), accepted values:
            ('newest', 'oldest', 'watchers'), API default: 'newest'
        :param int number: (optional), number of forks to return. Default: -1
            returns all forks
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Repository <Repository>`
        """
        url = self._build_url('forks', base_url=self._api)
        params = {}
        if sort in ('newest', 'oldest', 'watchers'):
            params = {'sort': sort}
        return self._iter(int(number), url, Repository, params, etag)

    def git_commit(self, sha):
        """Get a single (git) commit.

        :param str sha: (required), sha of the commit
        :returns: :class:`Commit <github3.git.Commit>` if successful,
            otherwise None
        """
        json = {}
        if sha:
            url = self._build_url('git', 'commits', sha, base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Commit, json)

    @requires_auth
    def hook(self, hook_id):
        """Get a single hook.

        :param int hook_id: (required), id of the hook
        :returns: :class:`Hook <github3.repos.hook.Hook>` if successful,
            otherwise None
        """
        json = None
        if int(hook_id) > 0:
            url = self._build_url('hooks', str(hook_id), base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Hook, json)

    @requires_auth
    def hooks(self, number=-1, etag=None):
        r"""Iterate over hooks registered on this repository.

        :param int number: (optional), number of hoks to return. Default: -1
            returns all hooks
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Hook <github3.repos.hook.Hook>`\ s
        """
        url = self._build_url('hooks', base_url=self._api)
        return self._iter(int(number), url, Hook, etag=etag)

    @requires_auth
    def ignore(self):
        """Ignore notifications from this repository for the user.

        .. versionadded:: 1.0

        This replaces ``Repository#set_subscription``.

        :returns: :class:`Subscription <github3.notifications.Subscription>`
        """
        url = self._build_url('subscription', base_url=self._api)
        json = self._json(self._put(url, data=dumps({'ignored': True})), 200)
        return self._instance_or_null(Subscription, json)

    @requires_auth
    def imported_issue(self, imported_issue_id):
        """Retrieve imported issue specified by imported issue id.

        :param int imported_issue_id: (required) id of imported issue
        :returns: :class:`Imported Issue <github3.repos.
            issue_import.ImportedIssue>`
        """
        url = self._build_url('import/issues', imported_issue_id,
                              base_url=self._api)
        data = self._get(url, headers=ImportedIssue.IMPORT_CUSTOM_HEADERS)
        json = self._json(data, 200)
        return self._instance_or_null(ImportedIssue, json)

    @requires_auth
    def imported_issues(self, number=-1, since=None, etag=None):
        """Retrieve the collection of imported issues via the API.

        See also: https://gist.github.com/jonmagic/5282384165e0f86ef105

        :param int number: (optional), number of imported issues to return.
            Default: -1 returns all branches
        :param since: (optional), Only imported issues after this date will
            be returned. This can be a ``datetime`` instance, ISO8601
            formatted date string, or a string formatted like so:
            ``2016-02-04`` i.e. ``%Y-%m-%d``
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`ImportedIssue <github3.repos.
            issue_import.ImportedIssue>`
        """

        data = {
            'since': timestamp_parameter(since)
        }

        self._remove_none(data)
        url = self._build_url('import/issues', base_url=self._api)

        return self._iter(int(number), url, ImportedIssue, etag=etag,
                          params=data,
                          headers=ImportedIssue.IMPORT_CUSTOM_HEADERS)

    @requires_auth
    def import_issue(self, title, body, created_at, assignee=None,
                     milestone=None, closed=None, labels=None, comments=None):
        """Import an issue into the repository.

        See also: https://gist.github.com/jonmagic/5282384165e0f86ef105

        :param string title: (required) Title of issue
        :param string body: (required) Body of issue
        :param timestamp created_at: (required) Creation timestamp
        :param string assignee: (optional) Username to assign issue to
        :param int milestone: (optional) Milestone ID
        :param boolean closed: (optional) Status of issue is Closed if True
        :param list labels: (optional) List of labels containing string names
        :param list comments: (optional) List of dictionaries which contain
            created_at and body attributes
        :returns: :class:`ImportedIssue <github3.repos.
            issue_import.ImportedIssue>`
        """

        issue = {
            'issue': {
                'title': title,
                'body': body,
                'created_at': created_at,
                'assignee': assignee,
                'milestone': milestone,
                'closed': closed,
                'labels': labels,
            },
            'comments': comments
        }

        self._remove_none(issue)
        self._remove_none(issue['issue'])
        url = self._build_url('import/issues', base_url=self._api)

        data = self._post(url, data=issue,
                          headers=ImportedIssue.IMPORT_CUSTOM_HEADERS)

        json = self._json(data, 200)
        return self._instance_or_null(ImportedIssue, json)

    def is_assignee(self, username):
        """Check if the user can be assigned an issue on this repository.

        :param username: name of the user to check
        :type username: str or :class:`User <github3.users.User>`
        :returns: :class:`bool`
        """
        if not username:
            return False
        url = self._build_url('assignees', str(username), base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    @requires_auth
    def is_collaborator(self, username):
        """Check to see if ``username`` is a collaborator on this repository.

        :param username: (required), login for the user
        :type username: str or :class:`User <github3.users.User>`
        :returns: bool -- True if successful, False otherwise
        """
        if not username:
            return False
        url = self._build_url('collaborators', str(username),
                              base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    def issue(self, number):
        """Get the issue specified by ``number``.

        :param int number: (required), number of the issue on this repository
        :returns: :class:`Issue <github3.issues.issue.Issue>` if successful,
            otherwise None
        """
        json = None
        if int(number) > 0:
            url = self._build_url('issues', str(number), base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Issue, json)

    def issue_events(self, number=-1, etag=None):
        r"""Iterate over issue events on this repository.

        :param int number: (optional), number of events to return. Default: -1
            returns all available events
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`IssueEvent <github3.issues.event.IssueEvent>`\ s
        """
        url = self._build_url('issues', 'events', base_url=self._api)
        return self._iter(int(number), url, IssueEvent, etag=etag)

    def issues(self, milestone=None, state=None, assignee=None, mentioned=None,
               labels=None, sort=None, direction=None, since=None, number=-1,
               etag=None):
        r"""Iterate over issues on this repo based upon parameters passed.

        .. versionchanged:: 0.9.0

            The ``state`` parameter now accepts 'all' in addition to 'open'
            and 'closed'.

        :param int milestone: (optional), 'none', or '*'
        :param str state: (optional), accepted values: ('all', 'open',
            'closed')
        :param str assignee: (optional), 'none', '*', or login name
        :param str mentioned: (optional), user's login name
        :param str labels: (optional), comma-separated list of labels, e.g.
            'bug,ui,@high'
        :param sort: (optional), accepted values:
            ('created', 'updated', 'comments', 'created')
        :param str direction: (optional), accepted values: ('asc', 'desc')
        :param since: (optional), Only issues after this date will
            be returned. This can be a ``datetime`` or an ``ISO8601`` formatted
            date string, e.g., 2012-05-20T23:10:27Z
        :type since: datetime or string
        :param int number: (optional), Number of issues to return.
            By default all issues are returned
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Issue <github3.issues.issue.Issue>`\ s
        """
        url = self._build_url('issues', base_url=self._api)

        params = repo_issue_params(milestone, state, assignee, mentioned,
                                   labels, sort, direction, since)

        return self._iter(int(number), url, Issue, params, etag)

    @requires_auth
    def key(self, id_num):
        """Get the specified deploy key.

        :param int id_num: (required), id of the key
        :returns: :class:`Key <github3.users.Key>` if successful, else None
        """
        json = None
        if int(id_num) > 0:
            url = self._build_url('keys', str(id_num), base_url=self._api)
            json = self._json(self._get(url), 200)
        return Key(json, self) if json else None

    @requires_auth
    def keys(self, number=-1, etag=None):
        r"""Iterate over deploy keys on this repository.

        :param int number: (optional), number of keys to return. Default: -1
            returns all available keys
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Key <github3.users.Key>`\ s
        """
        url = self._build_url('keys', base_url=self._api)
        return self._iter(int(number), url, Key, etag=etag)

    def label(self, name):
        """Get the label specified by ``name``.

        :param str name: (required), name of the label
        :returns: :class:`Label <github3.issues.label.Label>` if successful,
            else None
        """
        json = None
        if name:
            url = self._build_url('labels', name, base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Label, json)

    def labels(self, number=-1, etag=None):
        r"""Iterate over labels on this repository.

        :param int number: (optional), number of labels to return. Default: -1
            returns all available labels
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Label <github3.issues.label.Label>`\ s
        """
        url = self._build_url('labels', base_url=self._api)
        return self._iter(int(number), url, Label, etag=etag)

    def languages(self, number=-1, etag=None):
        """Iterate over the programming languages used in the repository.

        :param int number: (optional), number of languages to return. Default:
            -1 returns all used languages
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of tuples
        """
        url = self._build_url('languages', base_url=self._api)
        return self._iter(int(number), url, tuple, etag=etag)

    @requires_auth
    def latest_pages_build(self):
        """Get the build information for the most recent Pages build.

        :returns: :class:`PagesBuild <github3.repos.pages.PagesBuild>`
        """
        url = self._build_url('pages', 'builds', 'latest', base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(PagesBuild, json)

    def latest_release(self):
        """Get the latest release.

        Draft releases and prereleases are not returned by this endpoint.

        :returns: :class:`Release <github3.repos.release.Release>`
        """
        url = self._build_url('releases', 'latest', base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(Release, json)

    def license(self):
        """Get the contents of a license for the repo

        :returns: :class:`License <github3.licenses.License>`
        """
        url = self._build_url('license', base_url=self._api)
        json = self._json(self._get(url, headers=License.CUSTOM_HEADERS), 200)
        return self._instance_or_null(License, json)

    @requires_auth
    def mark_notifications(self, last_read=''):
        """Mark all notifications in this repository as read.

        :param str last_read: (optional), Describes the last point that
            notifications were checked. Anything updated since this time will
            not be updated. Default: Now. Expected in ISO 8601 format:
            ``YYYY-MM-DDTHH:MM:SSZ``. Example: "2012-10-09T23:39:01Z".
        :returns: bool
        """
        url = self._build_url('notifications', base_url=self._api)
        mark = {'read': True}
        if last_read:
            mark['last_read_at'] = last_read
        return self._boolean(self._put(url, data=dumps(mark)),
                             205, 404)

    @requires_auth
    def merge(self, base, head, message=''):
        """Perform a merge from ``head`` into ``base``.

        :param str base: (required), where you're merging into
        :param str head: (required), where you're merging from
        :param str message: (optional), message to be used for the commit
        :returns: :class:`RepoCommit <github3.repos.commit.RepoCommit>`
        """
        url = self._build_url('merges', base_url=self._api)
        data = {'base': base, 'head': head}
        if message:
            data['commit_message'] = message
        json = self._json(self._post(url, data=data), 201)
        return self._instance_or_null(RepoCommit, json)

    def milestone(self, number):
        """Get the milestone indicated by ``number``.

        :param int number: (required), unique id number of the milestone
        :returns: :class:`Milestone <github3.issues.milestone.Milestone>`
        """
        json = None
        if int(number) > 0:
            url = self._build_url('milestones', str(number),
                                  base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Milestone, json)

    def milestones(self, state=None, sort=None, direction=None, number=-1,
                   etag=None):
        r"""Iterate over the milestones on this repository.

        :param str state: (optional), state of the milestones, accepted
            values: ('open', 'closed')
        :param str sort: (optional), how to sort the milestones, accepted
            values: ('due_date', 'completeness')
        :param str direction: (optional), direction to sort the milestones,
            accepted values: ('asc', 'desc')
        :param int number: (optional), number of milestones to return.
            Default: -1 returns all milestones
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`Milestone <github3.issues.milestone.Milestone>`\ s
        """
        url = self._build_url('milestones', base_url=self._api)
        accepted = {'state': ('open', 'closed'),
                    'sort': ('due_date', 'completeness'),
                    'direction': ('asc', 'desc')}
        params = {'state': state, 'sort': sort, 'direction': direction}
        for (k, v) in list(params.items()):
            if not (v and (v in accepted[k])):  # e.g., '' or None
                del params[k]
        if not params:
            params = None
        return self._iter(int(number), url, Milestone, params, etag)

    def network_events(self, number=-1, etag=None):
        r"""Iterate over events on a network of repositories.

        :param int number: (optional), number of events to return. Default: -1
            returns all available events
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Event <github3.events.Event>`\ s
        """
        base = self._api.replace('repos', 'networks', 1)
        url = self._build_url('events', base_url=base)
        return self._iter(int(number), url, Event, etag)

    @requires_auth
    def notifications(self, all=False, participating=False, since=None,
                      number=-1, etag=None):
        r"""Iterate over the notifications for this repository.

        :param bool all: (optional), show all notifications, including ones
            marked as read
        :param bool participating: (optional), show only the notifications the
            user is participating in directly
        :param since: (optional), filters out any notifications updated
            before the given time. This can be a `datetime` or an `ISO8601`
            formatted date string, e.g., 2012-05-20T23:10:27Z
        :type since: datetime or string
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Thread <github3.notifications.Thread>`
        """
        url = self._build_url('notifications', base_url=self._api)
        params = {
            'all': str(all).lower(),
            'participating': str(participating).lower(),
            'since': timestamp_parameter(since)
        }
        self._remove_none(params)
        return self._iter(int(number), url, Thread, params, etag)

    @requires_auth
    def pages(self):
        """Get information about this repository's pages site.

        :returns: :class:`PagesInfo <github3.repos.pages.PagesInfo>`
        """
        url = self._build_url('pages', base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(PagesInfo, json)

    @requires_auth
    def pages_builds(self, number=-1, etag=None):
        """Iterate over pages builds of this repository.

        :returns: generator of :class:`PagesBuild
            <github3.repos.pages.PagesBuild>`
        """
        url = self._build_url('pages', 'builds', base_url=self._api)
        return self._iter(int(number), url, PagesBuild, etag=etag)

    def pull_request(self, number):
        """Get the pull request indicated by ``number``.

        :param int number: (required), number of the pull request.
        :returns: :class:`PullRequest <github3.pulls.PullRequest>`
        """
        json = None
        if int(number) > 0:
            url = self._build_url('pulls', str(number), base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(PullRequest, json)

    def pull_requests(self, state=None, head=None, base=None, sort='created',
                      direction='desc', number=-1, etag=None):
        r"""List pull requests on repository.

        .. versionchanged:: 0.9.0

            - The ``state`` parameter now accepts 'all' in addition to 'open'
              and 'closed'.

            - The ``sort`` parameter was added.

            - The ``direction`` parameter was added.

        :param str state: (optional), accepted values: ('all', 'open',
            'closed')
        :param str head: (optional), filters pulls by head user and branch
            name in the format ``user:ref-name``, e.g., ``seveas:debian``
        :param str base: (optional), filter pulls by base branch name.
            Example: ``develop``.
        :param str sort: (optional), Sort pull requests by ``created``,
            ``updated``, ``popularity``, ``long-running``. Default: 'created'
        :param str direction: (optional), Choose the direction to list pull
            requests. Accepted values: ('desc', 'asc'). Default: 'desc'
        :param int number: (optional), number of pulls to return. Default: -1
            returns all available pull requests
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`PullRequest <github3.pulls.PullRequest>`\ s
        """
        url = self._build_url('pulls', base_url=self._api)
        params = {}

        if state:
            state = state.lower()
            if state in ('all', 'open', 'closed'):
                params['state'] = state

        params.update(head=head, base=base, sort=sort, direction=direction)
        self._remove_none(params)
        return self._iter(int(number), url, PullRequest, params, etag)

    def readme(self):
        """Get the README for this repository.

        :returns: :class:`Contents <github3.repos.contents.Contents>`
        """
        url = self._build_url('readme', base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(Contents, json)

    def ref(self, ref):
        """Get a reference pointed to by ``ref``.

        The most common will be branches and tags. For a branch, you must
        specify 'heads/branchname' and for a tag, 'tags/tagname'. Essentially,
        the system should return any reference you provide it in the namespace,
        including notes and stashes (provided they exist on the server).

        :param str ref: (required)
        :returns: :class:`Reference <github3.git.Reference>`
        """
        json = None
        if ref:
            url = self._build_url('git', 'refs', ref, base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Reference, json)

    def refs(self, subspace='', number=-1, etag=None):
        r"""Iterate over references for this repository.

        :param str subspace: (optional), e.g. 'tags', 'stashes', 'notes'
        :param int number: (optional), number of refs to return. Default: -1
            returns all available refs
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Reference <github3.git.Reference>`\ s
        """
        if subspace:
            args = ('git', 'refs', subspace)
        else:
            args = ('git', 'refs')
        url = self._build_url(*args, base_url=self._api)
        return self._iter(int(number), url, Reference, etag=etag)

    def release(self, id):
        """Get a single release.

        :param int id: (required), id of release
        :returns: :class:`Release <github3.repos.release.Release>`
        """
        json = None
        if int(id) > 0:
            url = self._build_url('releases', str(id), base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Release, json)

    def release_from_tag(self, tag_name):
        """Get a release by tag name.

        release_from_tag() returns a release with specified tag
        while release() returns a release with specified release id

        :param str tag_name: (required) name of tag
        :returns: :class:`Release <github3.repos.release.Release>`
        """
        url = self._build_url('releases', 'tags', tag_name,
                              base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(Release, json)

    def releases(self, number=-1, etag=None):
        r"""Iterate over releases for this repository.

        :param int number: (optional), number of refs to return. Default: -1
            returns all available refs
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`Release <github3.repos.release.Release>`\ s
        """
        url = self._build_url('releases', base_url=self._api)
        iterator = self._iter(int(number), url, Release, etag=etag)
        iterator.headers.update(Release.CUSTOM_HEADERS)
        return iterator

    @requires_auth
    def remove_collaborator(self, username):
        """Remove collaborator ``username`` from the repository.

        :param username: (required), login name of the collaborator
        :type username: str or :class:`User <github3.users.User>`
        :returns: bool
        """
        if not username:
            return False

        url = self._build_url('collaborators', str(username),
                              base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)

    def stargazers(self, number=-1, etag=None):
        r"""List users who have starred this repository.

        :param int number: (optional), number of stargazers to return.
            Default: -1 returns all subscribers available
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <github3.users.User>`\ s
        """
        url = self._build_url('stargazers', base_url=self._api)
        return self._iter(int(number), url, User, etag=etag)

    def statuses(self, sha, number=-1, etag=None):
        r"""Iterate over the statuses for a specific SHA.

        .. warning::

            Deprecated in v1.0. Also deprecated upstream
            https://developer.github.com/v3/repos/statuses/

        :param str sha: SHA of the commit to list the statuses of
        :param int number: (optional), return up to number statuses. Default:
            -1 returns all available statuses.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Status <github3.repos.status.Status>`
        """
        url = ''
        if sha:
            url = self._build_url('statuses', sha, base_url=self._api)
        return self._iter(int(number), url, Status, etag=etag)

    @requires_auth
    def subscribe(self):
        """Subscribe the user to this repository's notifications.

        .. versionadded:: 1.0

        This replaces ``Repository#set_subscription``

        :param bool subscribed: (required), determines if notifications should
            be received from this repository.
        :param bool ignored: (required), determines if notifications should be
            ignored from this repository.
        :returns: :class:`Subscription <github3.notifications.Subscription>`
        """
        url = self._build_url('subscription', base_url=self._api)
        json = self._json(self._put(url, data=dumps({'subcribed': True})),
                          200)
        return self._instance_or_null(Subscription, json)

    def subscribers(self, number=-1, etag=None):
        r"""Iterate over users subscribed to this repository.

        :param int number: (optional), number of subscribers to return.
            Default: -1 returns all subscribers available
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`User <github3.users.User>`
        """
        url = self._build_url('subscribers', base_url=self._api)
        return self._iter(int(number), url, User, etag=etag)

    @requires_auth
    def subscription(self):
        """Return subscription for this Repository.

        :returns: :class:`Subscription <github3.notifications.Subscription>`
        """
        url = self._build_url('subscription', base_url=self._api)
        json = self._json(self._get(url), 200)
        return self._instance_or_null(Subscription, json)

    def tag(self, sha):
        """Get an annotated tag.

        http://learn.github.com/p/tagging.html

        :param str sha: (required), sha of the object for this tag
        :returns: :class:`Tag <github3.git.Tag>`
        """
        json = None
        if sha:
            url = self._build_url('git', 'tags', sha, base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Tag, json)

    def tags(self, number=-1, etag=None):
        r"""Iterate over tags on this repository.

        :param int number: (optional), return up to at most number tags.
            Default: -1 returns all available tags.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`RepoTag <github3.repos.tag.RepoTag>`\ s
        """
        url = self._build_url('tags', base_url=self._api)
        return self._iter(int(number), url, RepoTag, etag=etag)

    @requires_auth
    def teams(self, number=-1, etag=None):
        r"""Iterate over teams with access to this repository.

        :param int number: (optional), return up to number Teams. Default: -1
            returns all Teams.
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of :class:`Team <github3.orgs.Team>`\ s
        """
        from ..orgs import Team
        url = self._build_url('teams', base_url=self._api)
        return self._iter(int(number), url, Team, etag=etag)

    def tree(self, sha):
        """Get a tree.

        :param str sha: (required), sha of the object for this tree
        :returns: :class:`Tree <github3.git.Tree>`
        """
        json = None
        if sha:
            url = self._build_url('git', 'trees', sha, base_url=self._api)
            json = self._json(self._get(url), 200)
        return self._instance_or_null(Tree, json)

    def weekly_commit_count(self):
        """Retrieve the total commit counts.

        .. note:: All statistics methods may return a 202. If github3.py
            receives a 202 in this case, it will return an emtpy dictionary.
            You should give the API a moment to compose the data and then re
            -request it via this method.

        ..versionadded:: 0.7

        The dictionary returned has two entries: ``all`` and ``owner``. Each
        has a fifty-two element long list of commit counts. (Note: ``all``
        includes the owner.) ``d['all'][0]`` will be the oldest week,
        ``d['all'][51]`` will be the most recent.

        :returns: dict
        """
        url = self._build_url('stats', 'participation', base_url=self._api)
        resp = self._get(url)
        if resp and resp.status_code == 202:
            return {}
        json = self._json(resp, 200)
        if json and json.get('ETag'):
            del json['ETag']
        if json and json.get('Last-Modified'):
            del json['Last-Modified']
        return json


def repo_issue_params(milestone=None,
                      state=None,
                      assignee=None,
                      mentioned=None,
                      labels=None,
                      sort=None,
                      direction=None,
                      since=None,
                      number=-1,
                      etag=None):
    """Validate and filter issue method parameters in one place."""
    params = {'assignee': assignee, 'mentioned': mentioned}
    if milestone in ('*', 'none') or isinstance(milestone, int):
        params['milestone'] = milestone
    Repository._remove_none(params)
    params.update(
        issue_params(None, state, labels, sort, direction,
                     since)
    )
    return params



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

formatted_emojis = u'\x1b[35m  1.   \x1b[0mdolls \x1b[0m\n\x1b[35m  2.   \x1b[0mpalm_tree \x1b[0m\n\x1b[35m  3.   \x1b[0muk \x1b[0m\n\x1b[35m  4.   \x1b[0m100 \x1b[0m\n\x1b[35m  5.   \x1b[0mbaby_chick \x1b[0m\n\x1b[0m'  # NOQA



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

import pickle


class MockFeedParser(object):

    def parse(self, url):
        if url == 'user_feed':
            with open('tests/data/user_feed.p', 'rb') as fp:
                return pickle.load(fp)
        else:
            with open('tests/data/trends.p', 'rb') as fp:
                return pickle.load(fp)



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

import re

import six
import shlex
from prompt_toolkit.completion import Completion

from .completions import META_LOOKUP_GH


class TextUtils(object):
    """Utilities for parsing and matching text."""

    def find_matches(self, word, collection, fuzzy):
        """Find all matches in collection for word.

        :type word: str
        :param word: The word before the cursor.

        :type collection: iterable
        :param collection: A collection of words to match.

        :type fuzzy: bool
        :param fuzzy: Determines whether to use fuzzy matching.

        :rtype: generator
        :return: Yields an instance of `prompt_toolkit.completion.Completion`.
        """
        word = self._last_token(word).lower()
        matches = []
        for suggestion in self._find_collection_matches(word,
                                                        collection,
                                                        fuzzy):
            matches.append(suggestion)
        return matches

    def get_tokens(self, text):
        """Parse out all tokens.

        :type text: str
        :param text: A string to be split into tokens.

        :rtype: list
        :return: A list of strings for each word in the text.
        """
        if text is not None:
            text = text.strip()
            words = self._safe_split(text)
            return words
        return []

    def _last_token(self, text):
        """Find the last word in text.

        :type text: str
        :param text: A string to parse and obtain the last word.

        :rtype: str
        :return: The last word in the text.
        """
        if text is not None:
            text = text.strip()
            if len(text) > 0:
                word = self._safe_split(text)[-1]
                word = word.strip()
                return word
        return ''

    def _fuzzy_finder(self, text, collection, case_sensitive=True):
        """Customized fuzzy finder with optional case-insensitive matching.

        Adapted from: https://github.com/amjith/fuzzyfinder.

        :type text: str
        :param text: Input string entered by user.

        :type collection: iterable
        :param collection: collection of strings which will be filtered based
            on the input `text`.

        :type case_sensitive: bool
        :param case_sensitive: Determines whether the find will be case
            sensitive.

        :rtype: generator
        :return: Yields a list of suggestions narrowed down from `collections`
            using the `text` input.
        """
        suggestions = []
        if case_sensitive:
            pat = '.*?'.join(map(re.escape, text))
        else:
            pat = '.*?'.join(map(re.escape, text.lower()))
        regex = re.compile(pat)
        for item in collection:
            if case_sensitive:
                r = regex.search(item)
            else:
                r = regex.search(item.lower())
            if r:
                suggestions.append((len(r.group()), r.start(), item))

        return (z for _, _, z in sorted(suggestions))

    def _find_collection_matches(self, word, collection, fuzzy):
        """Yield all matching names in list.

        :type word: str
        :param word: The word before the cursor.

        :type collection: iterable
        :param collection: A collection of words to match.

        :type fuzzy: bool
        :param fuzzy: Determines whether to use fuzzy matching.

        :rtype: generator
        :return: Yields an instance of `prompt_toolkit.completion.Completion`.
        """
        word = word.lower()
        if fuzzy:
            for suggestion in self._fuzzy_finder(word,
                                                 collection,
                                                 case_sensitive=False):
                yield Completion(suggestion,
                                 -len(word),
                                 display_meta='display_meta')
        else:
            for name in sorted(collection):
                if name.lower().startswith(word) or not word:
                    display = None
                    display_meta = None
                    # print(len(words))
                    if name in META_LOOKUP_GH:
                        display_meta = META_LOOKUP_GH[name]
                    yield Completion(name,
                                     -len(word),
                                     display=display,
                                     display_meta=display_meta)

    def _shlex_split(self, text):
        """Wrapper for shlex, because it does not seem to handle unicode in 2.6.

        :type text: str
        :param text: A string to split.

        :rtype: list
        :return: A list that contains words for each split element of text.
        """
        if six.PY2:
            text = text.encode('utf-8')
        return shlex.split(text)

    def _safe_split(self, text):
        """Safely splits the input text.

        Shlex can't always split. For example, "\" crashes the completer.

        :type text: str
        :param text: A string to split.

        :rtype: list
        :return: A list that contains words for each split element of text.
        """
        try:
            words = self._shlex_split(text)
            return words
        except:
            return text

    def sanitize_if_none(self, text):
        """Sanitize text to ensure it is not None

        :type text:str
        :param text: String to sanitize.

        :rtype: str
        :return: String which is empty if None.
        """
        return text if text is not None else ''



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

from prompt_toolkit.completion import Completer, Completion

from .completions import COMPLETIONS_GH, SUBCOMMANDS
from .completions_git import META_LOOKUP_GIT, META_LOOKUP_GIT_EXTRAS
from .utils import TextUtils


class CompleterGitsome(Completer):
    """Gitsome prompt toolkit completer.

    :type text_utils: :class:`utils.TextUtils`
    :param text_utils: An instance of `utils.TextUtils`.

    :type fuzzy_match: bool
    :param fuzzy_match: Determines whether to use fuzzy matching.
        Currently, this is always set to False but can be enabled.
    """

    def __init__(self):
        self.fuzzy_match = False
        self.text_utils = TextUtils()

    def build_completions_with_meta(self, line, prefix, completions):
        """Build prompt_toolkit Completions with meta info.

        :type line: list
        :param line: The input text as a list of words.

        :type prefix: string
        :param prefix: The current word.

        :type completions: list
        :param completions: Completions to build meta info for.

        :rtype: list
        :return: Completions with meta info.
        """
        completions_with_meta = []
        tokens = line.split(' ')
        if tokens[0] != 'gh':
            for comp in completions:
                display = None
                display_meta = None
                if 'git' in line and comp.strip() in META_LOOKUP_GIT:
                    display_meta = META_LOOKUP_GIT[comp.strip()]
                elif 'git' in line and comp.strip() in META_LOOKUP_GIT_EXTRAS:
                    display_meta = META_LOOKUP_GIT_EXTRAS[comp.strip()]
                completions_with_meta.append(
                    Completion(comp,
                               -len(prefix),
                               display=display,
                               display_meta=display_meta))
        return completions_with_meta

    def completing_command(self, words, word_before_cursor):
        """Determine if we are currently completing the gh command.

        :type words: list
        :param words: The input text broken into word tokens.

        :type word_before_cursor: str
        :param word_before_cursor: The current word before the cursor,
            which might be one or more blank spaces.

        :rtype: bool
        :return: Specifies whether we are currently completing the gh command.
        """
        if len(words) == 1 and word_before_cursor != '':
            return True
        else:
            return False

    def completing_subcommand(self, words, word_before_cursor):
        """Determine if we are currently completing a subcommand.

        :type words: list
        :param words: The input text broken into word tokens.

        :type word_before_cursor: str
        :param word_before_cursor: The current word before the cursor,
            which might be one or more blank spaces.

        :rtype: bool
        :return: Specifies whether we are currently completing a subcommand.
        """
        if (len(words) == 1 and word_before_cursor == '') \
                or (len(words) == 2 and word_before_cursor != ''):
            return True
        else:
            return False

    def completing_arg(self, words, word_before_cursor):
        """Determine if we are currently completing an arg.

        :type words: list
        :param words: The input text broken into word tokens.

        :type word_before_cursor: str
        :param word_before_cursor: The current word before the cursor,
            which might be one or more blank spaces.

        :rtype: bool
        :return: Specifies whether we are currently completing an arg.
        """
        if (len(words) == 2 and word_before_cursor == '') \
                or (len(words) == 3 and word_before_cursor != ''):
            return True
        else:
            return False

    def completing_subcommand_option(self, words, word_before_cursor):
        """Determine if we are currently completing an option.

        :type words: list
        :param words: The input text broken into word tokens.

        :type word_before_cursor: str
        :param word_before_cursor: The current word before the cursor,
            which might be one or more blank spaces.

        :rtype: list
        :return: A list of options.
        """
        options = []
        for subcommand, args_opts in COMPLETIONS_GH.items():
            if subcommand in words and \
                (words[-2] == subcommand or
                    self.completing_subcommand_option_util(subcommand, words)):
                options.extend(COMPLETIONS_GH[subcommand]['opts'])
        return options

    def completing_subcommand_option_util(self, option, words):
        """Determine if we are currently completing an option.

        Called by completing_subcommand_option as a utility method.

        :type words: list
        :param words: The input text broken into word tokens.

        :type word_before_cursor: str
        :param word_before_cursor: The current word before the cursor,
            which might be one or more blank spaces.

        :rtype: bool
        :return: Specifies whether we are currently completing an option.
        """
        # Example: Return True for: gh view 1 --pag
        if len(words) > 3:
            if option in words:
                return True
        return False

    def arg_completions(self, words, word_before_cursor):
        """Generates arguments completions based on the input.

        :type words: list
        :param words: The input text broken into word tokens.

        :type word_before_cursor: str
        :param word_before_cursor: The current word before the cursor,
            which might be one or more blank spaces.

        :rtype: list
        :return: A list of completions.
        """
        if 'gh' not in words:
            return []
        for subcommand, args_opts in COMPLETIONS_GH.items():
            if subcommand in words:
                args = list(COMPLETIONS_GH[subcommand]['args'].keys())
                if not args:
                    # Some commands don't have args, complete options instead.
                    args = list(COMPLETIONS_GH[subcommand]['opts'].keys())
                return args if args else []
        return []

    def get_completions(self, document, _):
        """Get completions for the current scope.

        :type document: :class:`prompt_toolkit.Document`
        :param document: An instance of `prompt_toolkit.Document`.

        :type _: :class:`prompt_toolkit.completion.Completion`
        :param _: (Unused).

        :rtype: generator
        :return: Yields an instance of `prompt_toolkit.completion.Completion`.
        """
        word_before_cursor = document.get_word_before_cursor(WORD=True)
        words = self.text_utils.get_tokens(document.text)
        commands = []
        if len(words) == 0:
            return commands
        if self.completing_command(words, word_before_cursor):
            commands = ['gh']
        else:
            if 'gh' not in words:
                return commands
            if self.completing_subcommand(words, word_before_cursor):
                commands = list(SUBCOMMANDS.keys())
            else:
                if self.completing_arg(words, word_before_cursor):
                    commands = self.arg_completions(words, word_before_cursor)
                else:
                    commands = self.completing_subcommand_option(
                        words,
                        word_before_cursor)
        completions = self.text_utils.find_matches(
            word_before_cursor, commands, fuzzy=self.fuzzy_match)
        return completions



# -*- coding: utf-8 -*-
"""
github3.auths
=============

This module contains the Authorization object.

"""
from __future__ import unicode_literals

from .decorators import requires_basic_auth
from .models import GitHubCore


class Authorization(GitHubCore):

    """The :class:`Authorization <Authorization>` object.

    Two authorization instances can be checked like so::

        a1 == a2
        a1 != a2

    And is equivalent to::

        a1.id == a2.id
        a1.id != a2.id

    See also: http://developer.github.com/v3/oauth/#oauth-authorizations-api

    """

    def _update_attributes(self, auth):
        self._api = auth.get('url')
        #: Details about the application (name, url)
        self.app = auth.get('app', {})
        #: Returns the Authorization token
        self.token = auth.get('token', '')
        #: App name
        self.name = self.app.get('name', '')
        #: URL about the note
        self.note_url = auth.get('note_url') or ''
        #: Note about the authorization
        self.note = auth.get('note') or ''
        #: List of scopes this applies to
        self.scopes = auth.get('scopes', [])
        #: Unique id of the authorization
        self.id = auth.get('id', 0)
        #: datetime object representing when the authorization was created.
        self.created_at = self._strptime(auth.get('created_at'))
        #: datetime object representing when the authorization was updated.
        self.updated_at = self._strptime(auth.get('updated_at'))

    def _repr(self):
        return '<Authorization [{0}]>'.format(self.name)

    def _update(self, scopes_data, note, note_url):
        """Helper for add_scopes, replace_scopes, remove_scopes."""
        if note is not None:
            scopes_data['note'] = note
        if note_url is not None:
            scopes_data['note_url'] = note_url
        json = self._json(self._post(self._api, data=scopes_data), 200)

        if json:
            self._update_attributes(json)
            return True

        return False

    @requires_basic_auth
    def add_scopes(self, scopes, note=None, note_url=None):
        """Adds the scopes to this authorization.

        .. versionadded:: 1.0

        :param list scopes: Adds these scopes to the ones present on this
            authorization
        :param str note: (optional), Note about the authorization
        :param str note_url: (optional), URL to link to when the user views
            the authorization
        :returns: True if successful, False otherwise
        :rtype: bool
        """
        return self._update({'add_scopes': scopes}, note, note_url)

    @requires_basic_auth
    def delete(self):
        """Delete this authorization."""
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_basic_auth
    def remove_scopes(self, scopes, note=None, note_url=None):
        """Remove the scopes from this authorization.

        .. versionadded:: 1.0

        :param list scopes: Remove these scopes from the ones present on this
            authorization
        :param str note: (optional), Note about the authorization
        :param str note_url: (optional), URL to link to when the user views
            the authorization
        :returns: True if successful, False otherwise
        :rtype: bool
        """
        return self._update({'rm_scopes': scopes}, note, note_url)

    @requires_basic_auth
    def replace_scopes(self, scopes, note=None, note_url=None):
        """Replace the scopes on this authorization.

        .. versionadded:: 1.0

        :param list scopes: Use these scopes instead of the previous list
        :param str note: (optional), Note about the authorization
        :param str note_url: (optional), URL to link to when the user views
            the authorization
        :returns: True if successful, False otherwise
        :rtype: bool
        """
        return self._update({'scopes': scopes}, note, note_url)



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

COMPLETIONS_GH = {
    'configure': {
        'desc': "Configures gitsome.",
        'args': {},
        'opts': {
            '-e': 'flag (opt) configure GitHub Enterprise.',
            '--enterprise': 'flag (opt) configure GitHub Enterprise.',
        },
    },
    'create-comment': {
        'desc': 'Creates a comment on the given issue.',
        'args': {
            'octocat/Spoon-Knife/1': 'str (req) user/repo/issue_number combo.',
        },
        'opts': {
            '-t': 'see associated -- option for details.',
            '--text': 'str (req) comment text.',
        },
    },
    'create-issue': {
        'desc': 'Creates an issue.',
        'args': {
            'octocat/Spoon-Knife': 'str (req) user/repository combo.',
        },
        'opts': {
            '-t': 'see associated -- option for details.',
            '--issue_title': 'str (req) issue title.',
            '-d': 'str (opt) issue description.',
            '--issue_desc': 'str (opt) issue description.',
        },
    },
    'create-repo': {
        'desc': 'Creates a repository.',
        'args': {
            'Spoon-Knife': 'str (req) repository name.',
        },
        'opts': {
            '-d': 'str (opt) repo description',
            '--repo_desc': 'str (opt) repo description.',
            '-pr': 'flag (opt) create a private repo',
            '--private': 'flag (opt) create a private repo',
        },
    },
    'emails': {
        'desc': "Lists all the user's registered emails.",
        'args': {},
        'opts': {},
    },
    'emojis': {
        'desc': 'Lists all GitHub supported emojis.',
        'args': {},
        'opts': {
            '-p': 'flag (req) show results in a pager.',
            '--pager': 'flag (req) show results in a pager.',
        },
    },
    'feed': {
        'desc': "Lists all activity for the given user or repo, if called with no arg, shows the logged in user's feed.",
        'args': {
            'octocat/Hello-World --pager': "str (opt) user or user/repository combo, if blank, shows the logged in user's feed.",
        },
        'opts': {
            '-pr': 'flag (req) also show private events.',
            '--private': 'flag (req) also show private events.',
            '-p': 'flag (req) show results in a pager.',
            '--pager': 'flag (req) show results in a pager.',
        },
    },
    'followers': {
        'desc': 'Lists all followers and the total follower count.',
        'args': {
            'octocat': "str (req) the user's login id, if blank, shows logged in user's info.",
        },
        'opts': {
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'following': {
        'desc': 'Lists all followed users and the total followed count.',
        'args': {
            'octocat': "str (req) the user's login id, if blank, shows logged in user's info.",
        },
        'opts': {
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'gitignore-template': {
        'desc': 'Outputs the gitignore template for the given language.',
        'args': {
            'Python': 'str (req) the language-specific .gitignore.',
        },
        'opts': {},
    },
    'gitignore-templates': {
        'desc': 'Outputs all supported gitignore templates.',
        'args': {},
        'opts': {
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'issue': {
        'desc': 'Outputs detailed information about the given issue.',
        'args': {
            'octocat/Spoon-Knife/1': 'str (req) user/repo/issue_number combo.',
        },
        'opts': {},
    },
    'issues': {
        'desc': 'Lists all issues matching the filter.',
        'args': {},
        'opts': {
            '-f': 'str (opt) "assigned", "created", "mentioned", "subscribed" (default).',
            '--issue_filter': 'str (opt) "assigned", "created", "mentioned", "subscribed" (default).',
            '-s': 'str (opt) "all", "open" (default), "closed".',
            '--issue_state': 'str (opt) "all", "open" (default), "closed".',
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'license': {
        'desc': 'Outputs the license template for the given license.',
        'args': {
            'apache-2.0': 'str (req) the license name.',
        },
        'opts': {},
    },
    'licenses': {
        'desc': 'Outputs all supported license templates.',
        'args': {},
        'opts': {},
    },
    'me': {
        'desc': 'Lists information about the logged in user.',
        'args': {},
        'opts': {
            '-b': 'flag (opt) view profile in a browser instead of the terminal.',
            '--browser': 'flag (opt) view profile in a browser instead of the terminal.',
            '-t': 'see associated -- option for details.',
            '--text_avatar': 'flag (opt) view profile pic in plain text.',
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'notifications': {
        'desc': 'Lists all notifications.',
        'args': {},
        'opts': {
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'octo': {
        'desc': 'Outputs an Easter egg or the given message from Octocat.',
        'args': {
            '"Keep it logically awesome"': 'str (req) a message from Octocat, if empty, Octocat speaks an Easter egg.',
        },
        'opts': {},
    },
    'pull-request': {
        'desc': 'Outputs detailed information about the given pull request.',
        'args': {
            'octocat/Spoon-Knife/3': 'str (req) user/repo/pull_number combo.',
        },
        'opts': {},
    },
    'pull-requests': {
        'desc': 'Lists all pull requests.',
        'args': {},
        'opts': {
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'rate-limit': {
        'desc': 'Outputs the rate limit.',
        'args': {},
        'opts': {},
    },
    'repo': {
        'desc': 'Outputs detailed information about the given repo.',
        'args': {
            'octocat/Spoon-Knife': 'str (req) user/repository combo.',
        },
        'opts': {},
    },
    'repos': {
        'desc': 'Lists all repos matching the given filter.',
        'args': {
            '"foo bar optional filter"': 'str (opt) filters repos by name.',
        },
        'opts': {
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'search-issues': {
        'desc': 'Searches for all issues matching the given query.',
        'args': {
            '"foobarbaz in:title created:>=2015-01-01"': 'str (req) the search query.',
        },
        'opts': {
            '-s': 'str (opt) "stars", "forks", "updated", if blank, sorts by query best match.',
            '--sort': 'str (opt) "stars", "forks", "updated", if blank, sorts by query best match.',
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'search-repos': {
        'desc': 'Searches for all repos matching the given query.',
        'args': {
            '"created:>=2015-01-01 stars:>=1500 language:python"': 'str (req) the search query.',
        },
        'opts': {
            '-s': 'str (opt) "stars", "forks", "updated", if blank, sorts by query best match.',
            '--sort': 'str (opt) "stars", "forks", "updated", if blank, sorts by query best match.',
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'starred': {
        'desc': 'Outputs starred repos.',
        'args': {
            '"foo bar optional filter"': 'str (opt) filters repos by name.',
        },
        'opts': {
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'trending': {
        'desc': 'Lists trending repos for the given language.',
        'args': {
            'Overall': 'str (opt) the language filter for trending repos, if blank, the overall rankings are shown.',
        },
        'opts': {
            '-w': 'flag (opt) show the weekly trending repos.',
            '--weekly': 'flag (opt) show the weekly trending repos.',
            '-m': 'flag (opt) show the monthly trending repos.',
            '--monthly': 'flag (opt) show the monthly trending repos.',
            '-D': 'flag (opt) view trending devs.  Only valid with -b/--browser.',
            '--devs': 'flag (opt) view trending devs.  Only valid with -b/--browser.',
            '-b': 'flag (opt) view profile in a browser instead of the terminal.',
            '--browser': 'flag (opt) view profile in a browser instead of the terminal.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'user': {
        'desc': 'Lists information about the given user.',
        'args': {
            'github --pager': "str (req) the user's login id.",
        },
        'opts': {
            '-b': 'flag (opt) view profile in a browser instead of the terminal.',
            '--browser': 'flag (opt) view profile in a browser instead of the terminal.',
            '-t': 'see associated -- option for details.',
            '--text_avatar': 'flag (opt) view profile pic in plain text.',
            '-l': 'flag (opt) num items to show, defaults to 1000.',
            '--limit': 'flag (opt) num items to show, defaults to 1000.',
            '-p': 'flag (opt) show results in a pager.',
            '--pager': 'flag (opt) show results in a pager.',
        },
    },
    'view': {
        'desc': 'Views the given repo or issue index in the terminal or a browser.',
        'args': {
            '1': 'int (req) the 1-based index to view.',
        },
        'opts': {
            '-b': 'flag (opt) view profile in a browser instead of the terminal.',
            '--browser': 'flag (opt) view profile in a browser instead of the terminal.',
        },
    },
}
META_LOOKUP_GH = {
    '10': 'limit: int (opt) limits the posts displayed',
    '"(?i)(Python|Django)"': ('regex_query: string (opt) applies a regular '
                              'expression comment filter'),
    '1': 'index: int (req) views the post index',
    '"user"': 'user:string (req) shows info on the specified user',
    'gh': 'Git auto-completer with GitHub integration.',
}
SUBCOMMANDS = {}


def build_meta_lookups():
    for subcommand, args_opts in COMPLETIONS_GH.items():
        META_LOOKUP_GH.update({subcommand: COMPLETIONS_GH[subcommand]['desc']})
        SUBCOMMANDS.update({subcommand: COMPLETIONS_GH[subcommand]['desc']})
        for opt, meta in args_opts['opts'].items():
            META_LOOKUP_GH.update({opt: meta})
        for arg, meta in args_opts['args'].items():
            META_LOOKUP_GH.update({arg: meta})


build_meta_lookups()



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import unicode_literals
from __future__ import print_function

import click
from getpass import getpass
import os
import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning

from .compat import configparser
from .lib.github3 import authorize, enterprise_login, login
from .lib.github3.exceptions import AuthenticationFailed, UnprocessableEntity


class Config(object):
    """Gitsome config.

    :type api: :class:`github3.github.Github`
    :param api: An instance of github3.github.Github.

    :type clr_x: str
    :param clr_x: Various ansi color config colors to use for highlights.

    :type CONFIG: str
    :param CONFIG: The config file name.

    :type CONFIG_SECTION: str
    :param CONFIG_SECTION: The main config file section label.

    :type CONFIG_CLR_X: str
    :param CONFIG_CLR_X: Various ansi color config labels to use for highlights.

    :type CONFIG_ENTERPRISE_URL: str
    :param CONFIG_ENTERPRISE_URL: The GitHub Enterprise url.

    :type CONFIG_USER_LOGIN: str
    :param CONFIG_USER_LOGIN: The user login.

    :type CONFIG_USER_PASS: str
    :param CONFIG_USER_PASS: The user password.

    :type CONFIG_USER_TOKEN: str
    :param CONFIG_USER_TOKEN: The user token.

    :type CONFIG_USER_FEED: str
    :param CONFIG_USER_FEED: The user feed config.  This is the feed on
        https://github.com/ when logged in and requires the basic auth model,
        which doesn't work when logging in with tokens or 2FA.  This config
        listed the pre-signed url to access the feed.

    :type CONFIG_URL: str
    :param CONFIG_URL: The config file name that contains urls used in the
        `gh view` command.

    :type CONFIG_URL_SECTION: str
    :param CONFIG_URL_SECTION: The config file section that contains urls used
        in the `gh view  [url_index]` command.

    :type CONFIG_URL_LIST: str
    :param CONFIG_URL_LIST: The config containing a list of the last set of
        urls the user has seen, which allows the user to quickly access a repo
        url with the `gh view [url_index]` command.

    :type CONFIG_VERIFY_SSL: str
    :param CONFIG_VERIFY_SSL: Determines whether to verify SSL certs.

    :type enterprise_url: str
    :param enterprise_url: The GitHub Enterprise url.

    :type urls: list
    :param urls: The last set of urls the user has seen, which allows the user
        to quickly access a repo url with the gh view [url_index] command.

    :type user_login: str
    :param user_login: The user's login in ~/.gitsomeconfig.

    :type user_pass: str
    :param user_pass: The user's pass in ~/.gitsomeconfig.
        This is only stored for GitHub Enterprise users since using only a
        personal access token does not seem to be supported.

    :type user_token: str
    :param user_token: The user's token in ~/.gitsomeconfig.

    :type verify_ssl: bool
    :param verify_ssl: Determines whether to verify SSL certs.
    """

    CONFIG = '.gitsomeconfig'
    CONFIG_CLR_PRIMARY = 'clr_primary'
    CONFIG_CLR_SECONDARY = 'clr_secondary'
    CONFIG_CLR_TERTIARY = 'clr_tertiary'
    CONFIG_CLR_QUATERNARY = 'clr_quaternary'
    CONFIG_CLR_BOLD = 'clr_bold'
    CONFIG_CLR_CODE = 'clr_code'
    CONFIG_CLR_ERROR = 'clr_error'
    CONFIG_CLR_HEADER = 'clr_header'
    CONFIG_CLR_LINK = 'clr_link'
    CONFIG_CLR_LIST = 'clr_list'
    CONFIG_CLR_MESSAGE = 'clr_message'
    CONFIG_CLR_NUM_COMMENTS = 'clr_num_comments'
    CONFIG_CLR_NUM_POINTS = 'clr_num_points'
    CONFIG_CLR_TAG = 'clr_tag'
    CONFIG_CLR_TIME = 'clr_time'
    CONFIG_CLR_TITLE = 'clr_title'
    CONFIG_CLR_TOOLTIP = 'clr_tooltip'
    CONFIG_CLR_USER = 'clr_user'
    CONFIG_CLR_VIEW_LINK = 'clr_view_link'
    CONFIG_CLR_VIEW_INDEX = 'clr_view_index'
    CONFIG_SECTION = 'github'
    CONFIG_USER_LOGIN = 'user_login'
    CONFIG_USER_PASS = 'user_pass'
    CONFIG_USER_TOKEN = 'user_token'
    CONFIG_USER_FEED = 'user_feed'
    CONFIG_ENTERPRISE_URL = 'enterprise_url'
    CONFIG_VERIFY_SSL = 'verify_ssl'
    CONFIG_URL = '.gitsomeconfigurl'
    CONFIG_URL_SECTION = 'url'
    CONFIG_URL_LIST = 'url_list'
    CONFIG_AVATAR = '.gitsomeconfigavatar.png'

    def __init__(self):
        self.api = None
        self.user_login = None
        self.user_pass = None
        self.user_token = None
        self.user_feed = None
        self.enterprise_url = None
        self.verify_ssl = True
        self.urls = []
        self._init_colors()
        self.load_configs([
            self.load_config_colors,
        ])
        self.login = login
        self.authorize = authorize
        self.getpass = getpass

    def _init_colors(self):
        """Initialize colors to their defaults."""
        self.clr_primary = None
        self.clr_secondary = 'green'
        self.clr_tertiary = 'cyan'
        self.clr_quaternary = 'yellow'
        self.clr_bold = 'cyan'
        self.clr_code = 'cyan'
        self.clr_error = 'red'
        self.clr_header = 'yellow'
        self.clr_link = 'green'
        self.clr_list = 'cyan'
        self.clr_message = None
        self.clr_num_comments = 'green'
        self.clr_num_points = 'green'
        self.clr_tag = 'cyan'
        self.clr_time = 'yellow'
        self.clr_title = None
        self.clr_tooltip = None
        self.clr_user = 'cyan'
        self.clr_view_link = 'magenta'
        self.clr_view_index = 'magenta'

    def authenticate_cached_credentials(self, config, parser,
                                        enterprise_auth=enterprise_login):
        """Authenticate with the user's credentials in ~/.gitsomeconfig.

        :type config: str
        :param config: The config path.

        :type parser: :class:`ConfigParser.RawConfigParser`
        :param parser: An instance of `ConfigParser.RawConfigParser.
        """
        with open(config) as config_file:
            try:
                parser.read_file(config_file)
            except AttributeError:
                parser.readfp(config_file)
            self.user_login = self.load_config(
                parser=parser,
                cfg_label=self.CONFIG_USER_LOGIN)
            self.user_pass = self.load_config(
                parser=parser,
                cfg_label=self.CONFIG_USER_PASS)
            self.user_token = self.load_config(
                parser=parser,
                cfg_label=self.CONFIG_USER_TOKEN)
            self.enterprise_url = self.load_config(
                parser=parser,
                cfg_label=self.CONFIG_ENTERPRISE_URL)
            self.verify_ssl = self.load_config(
                parser=parser,
                cfg_label=self.CONFIG_VERIFY_SSL,
                boolean_config=True)
            self.user_feed = self.load_config(
                parser=parser,
                cfg_label=self.CONFIG_USER_FEED)
            if not self.verify_ssl:
                # The user has chosen not to verify SSL certs.
                # Disable warnings related to this option.
                requests.packages.urllib3.disable_warnings(
                    InsecureRequestWarning)
            login_kwargs = {
                'username': self.user_login,
                'two_factor_callback': self.request_two_factor_code,
            }
            if self.enterprise_url is not None:
                self.login = enterprise_auth
                login_kwargs.update({
                    'url': self.enterprise_url,
                    'verify': self.verify_ssl,
                })
                if self.user_token is not None:
                    login_kwargs.update({'token': self.user_token})
                elif self.user_pass is not None:
                    login_kwargs.update({'password': self.user_pass})
                else:
                    self.print_auth_error()
                    return
            else:
                login_kwargs.update({'token': self.user_token})
            self.api = self.login(**login_kwargs)

    def authenticate(self, enterprise=False,
                     enterprise_auth=enterprise_login, overwrite=False):
        """Log into GitHub.

        Adapted from https://github.com/sigmavirus24/github-cli.

        :type enterprise: bool
        :param enterprise: Determines whether to configure GitHub Enterprise.
            Default: False.

        :type overwrite: bool
        :param overwrite: indicates whether we cant to overwrite the current
            set of credentials.  Default: False.
        """
        if self.api is not None and not overwrite:
            return
        # Get the full path to the configuration file.
        config = self.get_github_config_path(self.CONFIG)
        parser = configparser.RawConfigParser()
        # Check to make sure the file exists and we are allowed to read it.
        # Skip if we want to overwrite the auth settings.
        if os.path.isfile(config) and os.access(config, os.R_OK | os.W_OK) and \
                not overwrite:
            with open(config) as config_file:
                try:
                    parser.read_file(config_file)
                except AttributeError:
                    parser.readfp(config_file)
            self.authenticate_cached_credentials(config, parser)
        else:
            # The file didn't exist or we don't have the correct permissions.
            login_kwargs = {
                'two_factor_callback': self.request_two_factor_code,
            }
            if enterprise:
                self.login = enterprise_auth
                while not self.enterprise_url:
                    self.enterprise_url = input('Enterprise URL: ')
                if click.confirm('Do you want to verify SSL certs?',
                                 default=True):
                    self.verify_ssl = True
                else:
                    self.verify_ssl = False
                login_kwargs.update({
                    'url': self.enterprise_url,
                    'verify': self.verify_ssl,
                })
            while not self.user_login:
                self.user_login = input('User Login: ')
            login_kwargs.update({'username': self.user_login})
            if click.confirm(('Do you want to log in with a password [Y] or '
                              'a personal access token [n]?'),
                             default=True):
                user_pass = None
                while not user_pass:
                    user_pass = self.getpass('Password: ')
                login_kwargs.update({'password': user_pass})
                try:
                    if not enterprise:
                        # Trade the user password for a personal access token.
                        # This does not seem to be available for Enterprise.
                        auth = self.authorize(
                            self.user_login,
                            user_pass,
                            scopes=['user', 'repo'],
                            note='gitsome',
                            note_url='https://github.com/donnemartin/gitsome',
                            two_factor_callback=self.request_two_factor_code
                        )
                        self.user_token = auth.token
                    else:
                        self.user_pass = user_pass
                except (UnprocessableEntity, AuthenticationFailed):
                    click.secho('Error creating token.',
                                fg=self.clr_error)
                    click.secho(('Visit the following page and verify you do '
                                 'not have an existing token named "gitsome":\n'
                                 '  https://github.com/settings/tokens\n'
                                 'If a token already exists, update your '
                                 '~/.gitsomeconfig file with your token:\n'
                                 '  user_token = TOKEN\n'
                                 'You can also generate a new token.'),
                                fg=self.clr_message)
                    self.print_auth_error()
                    return
            else:
                # The user has chosen to authenticate with a token.
                while not self.user_token:
                    self.user_token = input('Token: ')
                login_kwargs.update({'token': self.user_token})
            self.api = self.login(**login_kwargs)
            if self.user_feed:
                parser.set(self.CONFIG_SECTION,
                           self.CONFIG_USER_FEED,
                           self.user_feed)

    def check_auth(self):
        """Check if the current authorization is valid.

        This method uses the ratelimit_remaining api to check whether
        the currently authenticated user's credentials are valid without
        deducting from the rate limit.  The ratelimit_remaining api does not
        seem to be available for GitHub Enterprise.

        github3.py's method check_authorization seems to only work given
        an authorization created by a registered application.

        TODO: Determine a better way to check the authorization for
        GitHub Enterprise.

        :type enterprise: bool
        :param enterprise: Determines whether we are authenticating with
            GitHub Enterprise.
        """
        if self.enterprise_url is not None:
            return True
        try:
            if self.api is not None:
                # Throws AuthenticationFailed if invalid credentials but
                # does not deduct from the rate limit.
                self.api.ratelimit_remaining
                return True
            else:
                self.print_auth_error()
        except AuthenticationFailed:
            self.print_auth_error()
        return False

    def get_github_config_path(self, config_file_name):
        """Attempt to find the github config file.

        Adapted from https://github.com/sigmavirus24/github-cli.

        :type config_file_name: str
        :param config_file_name: The config file name.

        :rtype: str
        :return: The github config file path.
        """
        home = os.path.abspath(os.environ.get('HOME', ''))
        config_file_path = os.path.join(home, config_file_name)
        return config_file_path

    def load_config(self, parser, cfg_label, default=None,
                    color_config=False, boolean_config=False):
        """Load the specified config from ~/.gitsomeconfig.

        :type parser: :class:`ConfigParser.RawConfigParser`
        :param parser: An instance of `ConfigParser.RawConfigParser`.

        :type cfg_label: str
        :param cfg_label: The config label to load.

        :type default: str
        :param default: The default color if no color config exists.
            Default: None.

        :type color_config: bool
        :param color_config: Determines whether this is a color config.
            Default: False.

        :type boolean_config: bool
        :param boolean_config: Determines whether to load a boolean config.
            Default: False.
        """
        try:
            if boolean_config:
                cfg = parser.getboolean(self.CONFIG_SECTION, cfg_label)
            else:
                cfg = parser.get(self.CONFIG_SECTION, cfg_label)
                if color_config:
                    if cfg == 'none':
                        cfg = None
                    # Check if the user input a valid color.
                    # If invalid, this will throw a TypeError
                    click.style('', fg=cfg)
        except (TypeError, configparser.NoOptionError):
            return default
        return cfg

    def load_configs(self, config_funcs):
        """Load the specified config from ~/.gitsomeconfig.

        :type foo: list
        :param foo: The config methods to run.
        """
        config_file_path = self.get_github_config_path(self.CONFIG)
        parser = configparser.RawConfigParser()
        try:
            with open(config_file_path) as config_file:
                try:
                    parser.read_file(config_file)
                except AttributeError:
                    parser.readfp(config_file)
                for config_func in config_funcs:
                    config_func(parser)
        except IOError:
            # There might not be a cache yet, just silently return.
            return None

    def load_config_colors(self, parser):
        """Load the color config from ~/.gitsomeconfig.

        :type parser: :class:`ConfigParser.RawConfigParser`
        :param parser: An instance of `ConfigParser.RawConfigParser`.
        """
        self.load_colors(parser)

    def load_colors(self, parser):
        """Load all colors from ~/.gitsomeconfig.

        :type parser: :class:`ConfigParser.RawConfigParser`
        :param parser: An instance of `ConfigParser.RawConfigParser`.
        """
        self.clr_primary = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_PRIMARY,
            default=self.clr_primary,
            color_config=True)
        self.clr_secondary = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_SECONDARY,
            default=self.clr_secondary,
            color_config=True)
        self.clr_tertiary = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_TERTIARY,
            default=self.clr_tertiary,
            color_config=True)
        self.clr_quaternary = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_QUATERNARY,
            default=self.clr_quaternary,
            color_config=True)
        self.clr_bold = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_BOLD,
            default=self.clr_bold,
            color_config=True)
        self.clr_code = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_CODE,
            default=self.clr_code,
            color_config=True)
        self.clr_code = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_ERROR,
            default=self.clr_code,
            color_config=True)
        self.clr_header = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_HEADER,
            default=self.clr_header,
            color_config=True)
        self.clr_link = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_LINK,
            default=self.clr_link,
            color_config=True)
        self.clr_list = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_LIST,
            default=self.clr_list,
            color_config=True)
        self.clr_message = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_MESSAGE,
            default=self.clr_message,
            color_config=True)
        self.clr_num_comments = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_NUM_COMMENTS,
            default=self.clr_num_comments,
            color_config=True)
        self.clr_num_points = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_NUM_POINTS,
            default=self.clr_num_points,
            color_config=True)
        self.clr_tag = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_TAG,
            default=self.clr_tag,
            color_config=True)
        self.clr_time = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_TIME,
            default=self.clr_time,
            color_config=True)
        self.clr_title = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_TITLE,
            default=self.clr_title,
            color_config=True)
        self.clr_tooltip = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_TOOLTIP,
            default=self.clr_tooltip,
            color_config=True)
        self.clr_user = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_USER,
            default=self.clr_user,
            color_config=True)
        self.clr_view_link = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_VIEW_LINK,
            default=self.clr_view_link,
            color_config=True)
        self.clr_view_index = self.load_config(
            parser=parser,
            cfg_label=self.CONFIG_CLR_VIEW_INDEX,
            default=self.clr_view_index,
            color_config=True)

    def load_urls(self, view_in_browser):
        """Load the current set of urls from ~/.gitsomeconfigurl.

        :type view_in_browser: bool
        :param view_in_browser: Determines whether to view the urls in a
            browser.

        :rtype: list
        :return: Collection of urls.
        """
        config = self.get_github_config_path(self.CONFIG_URL)
        parser = configparser.RawConfigParser()
        with open(config) as config_file:
            try:
                parser.read_file(config_file)
            except AttributeError:
                parser.readfp(config_file)
            urls = parser.get(self.CONFIG_URL_SECTION,
                              self.CONFIG_URL_LIST)
            urls = urls.strip()
            excludes = ['[', ']', "'"]
            for exclude in excludes:
                urls = urls.replace(exclude, '')
                if not view_in_browser:
                    urls = urls.replace('https://github.com/', '')
            return urls.split(', ')

    def print_auth_error(self):
        """Print a message the authorization has failed."""
        click.secho('Authentication error.', fg=self.clr_error)
        click.secho(('Update your credentials in ~/.gitsomeconfig '
                     'or run:\n  gh configure'),
                    fg=self.clr_message)

    def prompt_news_feed(self):
        """Prompt the user to enter a news feed url."""
        if click.confirm(('No feed url detected.\n  Calling gh events without '
                          "an argument\n  displays the logged in user's "
                          'news feed.\nDo you want gitsome to track your '
                          'news feed?'),
                         default=True):
            click.secho(('Visit the following url while logged into GitHub:\n'
                         '  https://github.com\n'
                         'Enter the url found under "Subscribe to your '
                         'news feed".'),
                        fg=self.clr_message)
            self.user_feed = ''
            while not self.user_feed:
                self.user_feed = input('URL: ')

    def request_two_factor_code(self):
        """Request two factor authentication code.

        Callback if two factor authentication is requested.

        :rtype: str
        :return: The user input two factor authentication code.
        """
        code = ''
        while not code:
            code = input('Enter 2FA code: ')
        return code

    def save_config(self):
        """Saves the config to ~/.gitsomeconfig."""
        if self.check_auth():
            config = self.get_github_config_path(self.CONFIG)
            parser = configparser.RawConfigParser()
            parser.add_section(self.CONFIG_SECTION)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_USER_LOGIN,
                       self.user_login)
            if self.user_token is not None:
                parser.set(self.CONFIG_SECTION,
                           self.CONFIG_USER_TOKEN,
                           self.user_token)
            if self.user_feed is not None:
                parser.set(self.CONFIG_SECTION,
                           self.CONFIG_USER_FEED,
                           self.user_feed)
            if self.enterprise_url is not None:
                parser.set(self.CONFIG_SECTION,
                           self.CONFIG_ENTERPRISE_URL,
                           self.enterprise_url)
                if self.user_pass is not None:
                    parser.set(self.CONFIG_SECTION,
                               self.CONFIG_USER_PASS,
                               self.user_pass)
            else:
                parser.remove_option(self.CONFIG_SECTION,
                                     self.CONFIG_USER_PASS)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_VERIFY_SSL,
                       self.verify_ssl)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_PRIMARY,
                       self.clr_primary)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_SECONDARY,
                       self.clr_secondary)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_TERTIARY,
                       self.clr_tertiary)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_QUATERNARY,
                       self.clr_quaternary)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_BOLD,
                       self.clr_bold)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_CODE,
                       self.clr_code)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_ERROR,
                       self.clr_error)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_HEADER,
                       self.clr_header)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_LINK,
                       self.clr_link)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_LIST,
                       self.clr_list)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_MESSAGE,
                       self.clr_message)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_NUM_COMMENTS,
                       self.clr_num_comments)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_NUM_POINTS,
                       self.clr_num_points)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_TAG,
                       self.clr_tag)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_TIME,
                       self.clr_time)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_TITLE,
                       self.clr_title)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_TOOLTIP,
                       self.clr_tooltip)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_USER,
                       self.clr_user)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_VIEW_LINK,
                       self.clr_view_link)
            parser.set(self.CONFIG_SECTION,
                       self.CONFIG_CLR_VIEW_INDEX,
                       self.clr_view_index)
            with open(config, 'w+') as config_file:
                parser.write(config_file)

    def save_urls(self):
        """Save the current set of urls to ~/.gitsomeconfigurl."""
        config = self.get_github_config_path(self.CONFIG_URL)
        parser = configparser.RawConfigParser()
        try:
            parser.add_section(self.CONFIG_URL_SECTION)
        except configparser.DuplicateSectionError:
            pass
        parser.set(self.CONFIG_URL_SECTION, self.CONFIG_URL_LIST, self.urls)
        with open(config, 'w+') as config_file:
            parser.write(config_file)

    def show_bash_completions_info(self):
        """Show info on how to enable bash completions"""
        click.secho(('By default, gitsome looks at the following locations '
                     'to enable bash completions:\n'
                     '  https://github.com/donnemartin/gitsome/blob/master/xonsh/environ.py#L123-L130\n'  # NOQA
                     'If bash completions are not working for you, check out '
                     'the following link:\n'
                     '  https://github.com/donnemartin/gitsome#enabling-bash-completions'),  # NOQA
                    fg=self.clr_message)



"""Xonsh hooks into bash completions."""
import builtins

import xonsh.platform as xp
from xonsh.completers.path import _quote_paths
from xonsh.completers.bash_completion import bash_completions


def complete_from_bash(prefix, line, begidx, endidx, ctx):
    """Completes based on results from BASH completion."""
    env = builtins.__xonsh__.env.detype()
    paths = builtins.__xonsh__.env.get("BASH_COMPLETIONS", ())
    command = xp.bash_command()
    return bash_completions(
        prefix,
        line,
        begidx,
        endidx,
        env=env,
        paths=paths,
        command=command,
        quote_paths=_quote_paths,
    )



# -*- coding: utf-8 -*-

# Copyright 2015 Donne Martin. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

formatted_org = u'To view the avatar in your terminal, install the Python Image Library.\n\x1b[0m\x1b[0muser2\n\x1b[0m\x1b[32mbigcorp\n\x1b[0m\x1b[32minterwebz\n\x1b[0m\x1b[32muser2@foo.com\n\x1b[0m\x1b[36mOrganization\n\n\x1b[0mNo results found\x1b[0m'
formatted_user = u'To view the avatar in your terminal, install the Python Image Library.\n\x1b[0m\x1b[0muser1\n\x1b[0m\x1b[32mbigcorp\n\x1b[0m\x1b[32minterwebz\n\x1b[0m\x1b[32muser1@foo.com\n\x1b[0m\x1b[36mFollowers: 0 | \x1b[0m\x1b[36mFollowing: 0\n\n\x1b[0m\x1b[35m  1.   \x1b[0mrepo1 \x1b[0m()\x1b[0m\n\x1b[32m        Stars: 1      \x1b[0m\x1b[36mForks: 1      \x1b[0m\x1b[33mUpdated: just now \x1b[0m\n  View the page for \x1b[0m\x1b[35m1 through \x1b[0m\x1b[35m1\x1b[0m with the following command:\n\x1b[0m\x1b[35m    gh view [#] \x1b[0moptional: [-b/--browser] [--help]\n\x1b[0m\x1b[0m'
formatted_users = u'\x1b[35m  1.   \x1b[0mfoo1 \x1b[0m\n\x1b[35m  2.   \x1b[0mfoo2 \x1b[0m\n\x1b[35m  3.   \x1b[0mfoo3 \x1b[0m\n  View the page for \x1b[0m\x1b[35m1 through \x1b[0m\x1b[35m3\x1b[0m with the following command:\n\x1b[0m\x1b[35m    gh view [#] \x1b[0moptional: [-b/--browser] [--help]\n\x1b[0m\x1b[0m'



# -*- coding: utf-8 -*-
"""
github3.repos.hook
==================

This module contains only the Hook object for GitHub's Hook API.

"""
from __future__ import unicode_literals

from json import dumps
from ..decorators import requires_auth
from ..models import GitHubCore


class Hook(GitHubCore):
    """The :class:`Hook <Hook>` object. This handles the information returned
    by GitHub about hooks set on a repository.

    Two hook instances can be checked like so::

        h1 == h2
        h1 != h2

    And is equivalent to::

        h1.id == h2.id
        h1.id != h2.id

    See also: http://developer.github.com/v3/repos/hooks/
    """
    def _update_attributes(self, hook, session=None):
        self._api = hook.get('url', '')
        #: datetime object representing when this hook was last updated.
        self.updated_at = self._strptime(hook.get('updated_at'))
        #: datetime object representing the date the hook was created.
        self.created_at = self._strptime(hook.get('created_at'))
        #: The name of the hook.
        self.name = hook.get('name')
        #: Events which trigger the hook.
        self.events = hook.get('events')
        #: Whether or not this Hook is marked as active on GitHub
        self.active = hook.get('active')
        #: Dictionary containing the configuration for the Hook.
        self.config = hook.get('config')
        #: Unique id of the hook.
        self.id = hook.get('id')

    def _repr(self):
        return '<Hook [{0}]>'.format(self.name)

    @requires_auth
    def delete(self):
        """Delete this hook.

        :returns: bool
        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(self, config={}, events=[], add_events=[], rm_events=[],
             active=True):
        """Edit this hook.

        :param dict config: (optional), key-value pairs of settings for this
            hook
        :param list events: (optional), which events should this be triggered
            for
        :param list add_events: (optional), events to be added to the list of
           events that this hook triggers for
        :param list rm_events: (optional), events to be removed from the list
            of events that this hook triggers for
        :param bool active: (optional), should this event be active
        :returns: bool
        """
        data = {'config': config, 'active': active}
        if events:
            data['events'] = events

        if add_events:
            data['add_events'] = add_events

        if rm_events:
            data['remove_events'] = rm_events

        json = self._json(self._patch(self._api, data=dumps(data)), 200)

        if json:
            self._update_attributes(json)
            return True

        return False

    @requires_auth
    def ping(self):
        """Ping this hook.

        :returns: bool
        """
        url = self._build_url('pings', base_url=self._api)
        return self._boolean(self._post(url), 204, 404)

    @requires_auth
    def test(self):
        """Test this hook

        :returns: bool
        """
        url = self._build_url('tests', base_url=self._api)
        return self._boolean(self._post(url), 204, 404)



__version__ = "0.8.10"


# amalgamate exclude jupyter_kernel parser_table parser_test_table pyghooks
# amalgamate exclude winutils wizard pytest_plugin fs macutils pygments_cache
# amalgamate exclude jupyter_shell
import os as _os

if _os.getenv("XONSH_DEBUG", ""):
    pass
else:
    import sys as _sys

    try:
        from xonsh import __amalgam__

        completer = __amalgam__
        _sys.modules["xonsh.completer"] = __amalgam__
        contexts = __amalgam__
        _sys.modules["xonsh.contexts"] = __amalgam__
        lazyasd = __amalgam__
        _sys.modules["xonsh.lazyasd"] = __amalgam__
        lazyjson = __amalgam__
        _sys.modules["xonsh.lazyjson"] = __amalgam__
        color_tools = __amalgam__
        _sys.modules["xonsh.color_tools"] = __amalgam__
        platform = __amalgam__
        _sys.modules["xonsh.platform"] = __amalgam__
        pretty = __amalgam__
        _sys.modules["xonsh.pretty"] = __amalgam__
        codecache = __amalgam__
        _sys.modules["xonsh.codecache"] = __amalgam__
        lazyimps = __amalgam__
        _sys.modules["xonsh.lazyimps"] = __amalgam__
        parser = __amalgam__
        _sys.modules["xonsh.parser"] = __amalgam__
        tokenize = __amalgam__
        _sys.modules["xonsh.tokenize"] = __amalgam__
        tools = __amalgam__
        _sys.modules["xonsh.tools"] = __amalgam__
        ansi_colors = __amalgam__
        _sys.modules["xonsh.ansi_colors"] = __amalgam__
        ast = __amalgam__
        _sys.modules["xonsh.ast"] = __amalgam__
        commands_cache = __amalgam__
        _sys.modules["xonsh.commands_cache"] = __amalgam__
        diff_history = __amalgam__
        _sys.modules["xonsh.diff_history"] = __amalgam__
        events = __amalgam__
        _sys.modules["xonsh.events"] = __amalgam__
        foreign_shells = __amalgam__
        _sys.modules["xonsh.foreign_shells"] = __amalgam__
        jobs = __amalgam__
        _sys.modules["xonsh.jobs"] = __amalgam__
        jsonutils = __amalgam__
        _sys.modules["xonsh.jsonutils"] = __amalgam__
        lexer = __amalgam__
        _sys.modules["xonsh.lexer"] = __amalgam__
        openpy = __amalgam__
        _sys.modules["xonsh.openpy"] = __amalgam__
        style_tools = __amalgam__
        _sys.modules["xonsh.style_tools"] = __amalgam__
        xontribs = __amalgam__
        _sys.modules["xonsh.xontribs"] = __amalgam__
        dirstack = __amalgam__
        _sys.modules["xonsh.dirstack"] = __amalgam__
        inspectors = __amalgam__
        _sys.modules["xonsh.inspectors"] = __amalgam__
        proc = __amalgam__
        _sys.modules["xonsh.proc"] = __amalgam__
        shell = __amalgam__
        _sys.modules["xonsh.shell"] = __amalgam__
        timings = __amalgam__
        _sys.modules["xonsh.timings"] = __amalgam__
        xonfig = __amalgam__
        _sys.modules["xonsh.xonfig"] = __amalgam__
        base_shell = __amalgam__
        _sys.modules["xonsh.base_shell"] = __amalgam__
        environ = __amalgam__
        _sys.modules["xonsh.environ"] = __amalgam__
        tracer = __amalgam__
        _sys.modules["xonsh.tracer"] = __amalgam__
        readline_shell = __amalgam__
        _sys.modules["xonsh.readline_shell"] = __amalgam__
        replay = __amalgam__
        _sys.modules["xonsh.replay"] = __amalgam__
        aliases = __amalgam__
        _sys.modules["xonsh.aliases"] = __amalgam__
        dumb_shell = __amalgam__
        _sys.modules["xonsh.dumb_shell"] = __amalgam__
        built_ins = __amalgam__
        _sys.modules["xonsh.built_ins"] = __amalgam__
        execer = __amalgam__
        _sys.modules["xonsh.execer"] = __amalgam__
        imphooks = __amalgam__
        _sys.modules["xonsh.imphooks"] = __amalgam__
        main = __amalgam__
        _sys.modules["xonsh.main"] = __amalgam__
        del __amalgam__
    except ImportError:
        pass
    del _sys
del _os
# amalgamate end



# -*- coding: utf-8 -*-
"""
github3.gists.file
------------------

Module containing the logic for the GistFile object.
"""
from __future__ import unicode_literals

from ..models import GitHubCore


class GistFile(GitHubCore):

    """This represents the file object returned by interacting with gists.

    It stores the raw url of the file, the file name, language, size and
    content.

    """

    def _update_attributes(self, attributes):
        #: The raw URL for the file at GitHub.
        self.raw_url = attributes.get('raw_url')
        #: The name of the file.
        self.filename = attributes.get('filename')
        #: The name of the file.
        self.name = attributes.get('filename')
        #: The language associated with the file.
        self.language = attributes.get('language')
        #: The size of the file.
        self.size = attributes.get('size')
        #: The content of the file.
        self.original_content = attributes.get('content')

    def _repr(self):
        return '<Gist File [{0}]>'.format(self.name)

    def content(self):
        """Retrieve contents of file from key 'raw_url' if there is no
        'content' key in Gist object.
        """
        resp = self._get(self.raw_url)
        if self._boolean(resp, 200, 404):
            return resp.content
        return None



# -*- coding: utf-8 -*-
"""Implements the xonsh parser for Python v3.6."""
import xonsh.ast as ast
from xonsh.parsers.v35 import Parser as ThreeFiveParser
from xonsh.parsers.base import store_ctx, ensure_has_elts, lopen_loc


class Parser(ThreeFiveParser):
    """A Python v3.6 compliant parser for the xonsh language."""

    def p_comp_for(self, p):
        """comp_for : FOR exprlist IN or_test comp_iter_opt"""
        targs, it, p5 = p[2], p[4], p[5]
        if len(targs) == 1:
            targ = targs[0]
        else:
            targ = ensure_has_elts(targs)
        store_ctx(targ)
        # only difference with base should be the is_async=0
        comp = ast.comprehension(target=targ, iter=it, ifs=[], is_async=0)
        comps = [comp]
        p0 = {"comps": comps}
        if p5 is not None:
            comps += p5.get("comps", [])
            comp.ifs += p5.get("if", [])
        p[0] = p0

    def p_expr_stmt_annassign(self, p):
        """expr_stmt : testlist_star_expr COLON test EQUALS test"""
        p1 = p[1][0]
        lineno, col = lopen_loc(p1)
        if len(p[1]) > 1 or not isinstance(p1, ast.Name):
            loc = self.currloc(lineno, col)
            self._parse_error("only single target can be annotated", loc)
        store_ctx(p1)
        p[0] = ast.AnnAssign(
            target=p1,
            annotation=p[3],
            value=p[5],
            simple=1,
            lineno=lineno,
            col_offset=col,
        )



# -*- coding: utf-8 -*-
"""
github3.gists.gist
==================

This module contains the Gist class alone for simplicity.

"""
from __future__ import unicode_literals

from json import dumps
from ..models import GitHubCore
from ..decorators import requires_auth
from .comment import GistComment
from .file import GistFile
from .history import GistHistory
from ..users import User


class Gist(GitHubCore):

    """This object holds all the information returned by Github about a gist.

    With it you can comment on or fork the gist (assuming you are
    authenticated), edit or delete the gist (assuming you own it).  You can
    also "star" or "unstar" the gist (again assuming you have authenticated).

    Two gist instances can be checked like so::

        g1 == g2
        g1 != g2

    And is equivalent to::

        g1.id == g2.id
        g1.id != g2.id

    See also: http://developer.github.com/v3/gists/

    """

    def _update_attributes(self, data):
        #: Number of comments on this gist
        self.comments_count = data.get('comments', 0)

        #: Unique id for this gist.
        self.id = '{0}'.format(data.get('id', ''))

        #: Description of the gist
        self.description = data.get('description', '')

        # e.g. https://api.github.com/gists/1
        self._api = data.get('url', '')

        #: URL of this gist at Github, e.g., https://gist.github.com/1
        self.html_url = data.get('html_url')
        #: Boolean describing if the gist is public or private
        self.public = data.get('public')

        self._forks = data.get('forks', [])

        #: Git URL to pull this gist, e.g., git://gist.github.com/1.git
        self.git_pull_url = data.get('git_pull_url', '')

        #: Git URL to push to gist, e.g., git@gist.github.com/1.git
        self.git_push_url = data.get('git_push_url', '')

        #: datetime object representing when the gist was created.
        self.created_at = self._strptime(data.get('created_at'))

        #: datetime object representing the last time this gist was updated.
        self.updated_at = self._strptime(data.get('updated_at'))

        owner = data.get('owner')
        #: :class:`User <github3.users.User>` object representing the owner of
        #: the gist.
        self.owner = User(owner, self) if owner else None

        self._files = [GistFile(data['files'][f]) for f in data['files']]

        #: History of this gist, list of
        #: :class:`GistHistory <github3.gists.history.GistHistory>`
        self.history = [GistHistory(h, self) for h in data.get('history', [])]

        # New urls

        #: Comments URL (not a template)
        self.comments_url = data.get('comments_url', '')

        #: Commits URL (not a template)
        self.commits_url = data.get('commits_url', '')

        #: Forks URL (not a template)
        self.forks_url = data.get('forks_url', '')

        #: Whether the content of this Gist has been truncated or not
        self.truncated = data.get('truncated')

    def __str__(self):
        return self.id

    def _repr(self):
        return '<Gist [{0}]>'.format(self.id)

    @requires_auth
    def create_comment(self, body):
        """Create a comment on this gist.

        :param str body: (required), body of the comment
        :returns: :class:`GistComment <github3.gists.comment.GistComment>`

        """
        json = None
        if body:
            url = self._build_url('comments', base_url=self._api)
            json = self._json(self._post(url, data={'body': body}), 201)
        return self._instance_or_null(GistComment, json)

    @requires_auth
    def delete(self):
        """Delete this gist.

        :returns: bool -- whether the deletion was successful

        """
        return self._boolean(self._delete(self._api), 204, 404)

    @requires_auth
    def edit(self, description='', files={}):
        """Edit this gist.

        :param str description: (optional), description of the gist
        :param dict files: (optional), files that make up this gist; the
            key(s) should be the file name(s) and the values should be another
            (optional) dictionary with (optional) keys: 'content' and
            'filename' where the former is the content of the file and the
            latter is the new name of the file.
        :returns: bool -- whether the edit was successful

        """
        data = {}
        json = None
        if description:
            data['description'] = description
        if files:
            data['files'] = files
        if data:
            json = self._json(self._patch(self._api, data=dumps(data)), 200)
        if json:
            self._update_attributes(json)
            return True
        return False

    @requires_auth
    def fork(self):
        """Fork this gist.

        :returns: :class:`Gist <Gist>` if successful, ``None`` otherwise

        """
        url = self._build_url('forks', base_url=self._api)
        json = self._json(self._post(url), 201)
        return self._instance_or_null(Gist, json)

    @requires_auth
    def is_starred(self):
        """Check to see if this gist is starred by the authenticated user.

        :returns: bool -- True if it is starred, False otherwise

        """
        url = self._build_url('star', base_url=self._api)
        return self._boolean(self._get(url), 204, 404)

    def comments(self, number=-1, etag=None):
        """Iterate over comments on this gist.

        :param int number: (optional), number of comments to iterate over.
            Default: -1 will iterate over all comments on the gist
        :param str etag: (optional), ETag from a previous request to the same
            endpoint
        :returns: generator of
            :class:`GistComment <github3.gists.comment.GistComment>`

        """
        url = self._build_url('comments', base_url=self._api)
        return self._iter(int(number), url, GistComment, etag=etag)

    def commits(self, number=-1, etag=None):
        """Iterate over the commits on this gist.

        These commits will be requested from the API and should be the same as
        what is in ``Gist.history``.

        .. versionadded:: 0.6

        .. versionchanged:: 0.9

            Added param ``etag``.

        :param int number: (optional), number of commits to iterate over.
            Default: -1 will iterate over all commits associated with this
            gist.
        :param str etag: (optional), ETag from a previous request to this
            endpoint.
        :returns: generator of
            :class:`GistHistory <github3.gists.history.GistHistory>`

        """
        url = self._build_url('commits', base_url=self._api)
        return self._iter(int(number), url, GistHistory)

    def files(self):
        """Iterator over the files stored in this gist.

        :returns: generator of :class`GistFile <github3.gists.file.GistFile>`

        """
        return iter(self._files)

    def forks(self, number=-1, etag=None):
        """Iterator of forks of this gist.

        .. versionchanged:: 0.9

            Added params ``number`` and ``etag``.

        :param int number: (optional), number of forks to iterate over.
            Default: -1 will iterate over all forks of this gist.
        :param str etag: (optional), ETag from a previous request to this
            endpoint.
        :returns: generator of :class:`Gist <Gist>`

        """
        url = self._build_url('forks', base_url=self._api)
        return self._iter(int(number), url, Gist, etag=etag)

    @requires_auth
    def star(self):
        """Star this gist.

        :returns: bool -- True if successful, False otherwise

        """
        url = self._build_url('star', base_url=self._api)
        return self._boolean(self._put(url), 204, 404)

    @requires_auth
    def unstar(self):
        """Un-star this gist.

        :returns: bool -- True if successful, False otherwise

        """
        url = self._build_url('star', base_url=self._api)
        return self._boolean(self._delete(url), 204, 404)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from ..models import GitHubCore
from ..users import User


class UserSearchResult(GitHubCore):
    def _update_attributes(self, data):
        result = data.copy()
        #: Score of this search result
        self.score = result.pop('score')
        #: Text matches
        self.text_matches = result.pop('text_matches', [])
        #: User object matching the search
        self.user = User(result, self)

    def _repr(self):
        return '<UserSearchResult [{0}]>'.format(self.user)



# -*- coding: utf-8 -*-
"""
github3
=======

See http://github3.readthedocs.org/ for documentation.

:copyright: (c) 2012-2016 by Ian Cordasco
:license: Modified BSD, see LICENSE for more details

"""

from .__about__ import (
    __package_name__, __title__, __author__, __author_email__,
    __license__, __copyright__, __version__, __version_info__,
    __url__,
)
from .api import (
    authorize, login, enterprise_login, emojis, gist, gitignore_template,
    create_gist, issue, markdown, octocat, organization, pull_request,
    followers_of, followed_by, public_gists, gists_by, issues_on,
    gitignore_templates, all_repositories, all_users, all_events,
    organizations_with, repositories_by, starred_by, subscriptions_for,
    rate_limit, repository, search_code, search_repositories, search_users,
    search_issues, user, zen
)
from .github import GitHub, GitHubEnterprise, GitHubStatus
from .exceptions import (
    BadRequest, AuthenticationFailed, ForbiddenError, GitHubError,
    MethodNotAllowed, NotFoundError, ServerError, NotAcceptable,
    UnprocessableEntity
)

__all__ = (
    'AuthenticationFailed',  'BadRequest', 'ForbiddenError', 'GitHub',
    'GitHubEnterprise', 'GitHubError', 'GitHubStatus',
    'MethodNotAllowed', 'NotAcceptable', 'NotFoundError', 'ServerError',
    'UnprocessableEntity', 'authorize', 'login', 'enterprise_login', 'emojis',
    'gist', 'gitignore_template', 'create_gist', 'issue', 'markdown',
    'octocat', 'organization', 'pull_request', 'followers_of', 'followed_by',
    'public_gists', 'gists_by', 'issues_on', 'gitignore_templates',
    'all_repositories', 'all_users', 'all_events', 'organizations_with',
    'repositories_by', 'starred_by', 'subscriptions_for', 'rate_limit',
    'repository', 'search_code', 'search_repositories', 'search_users',
    'search_issues', 'user', 'zen',
    # Metadata attributes
    '__package_name__', '__title__', '__author__', '__author_email__',
    '__license__', '__copyright__', '__version__', '__version_info__',
    '__url__',
)
